# coding: utf-8

import math

import cv2
import numpy as np
from typing import Tuple

from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.point_location import PointLocation


# to rotate image & objects wrt source img center
# output image will contain all 'pixels' from source img
class ImageRotator:
    """Rotates images and geometry figures around image center by a given angle."""
    # to get rect with max 'coloured' area in rotated img
    def _calc_inner_crop(self):
        """
        Given a rectangle of self.src_imsize HxW that has been rotated by
        self.angle_degrees_ccw (in degrees), computes the location of the
        largest possible axis-aligned rectangle within the rotated rectangle.
        """

        # TODO This needs significant streamlinig.
        a_ccw = np.deg2rad(self.angle_degrees_ccw)
        quadrant = math.floor(a_ccw / (math.pi / 2)) & 3
        sign_alpha = a_ccw if ((quadrant & 1) == 0) else math.pi - a_ccw
        alpha = (sign_alpha % math.pi + math.pi) % math.pi

        h, w = self.src_imsize
        bb_w = w * math.cos(alpha) + h * math.sin(alpha)
        bb_h = w * math.sin(alpha) + h * math.cos(alpha)

        gamma = math.atan2(bb_w, bb_w) if (w < h) else math.atan2(bb_w, bb_w)

        delta = math.pi - alpha - gamma

        length = h if (w < h) else w

        d = length * math.cos(alpha)
        a = d * math.sin(alpha) / math.sin(delta)

        y = a * math.cos(gamma)
        x = y * math.tan(gamma)

        largest_w, largest_h = bb_w - 2 * x, bb_h - 2 * y

        new_h, new_w = self.new_imsize
        left = round((new_w - largest_w) * 0.5)
        right = round((new_w + largest_w) * 0.5)
        top = round((new_h - largest_h) * 0.5)
        bottom = round((new_h + largest_h) * 0.5)
        some_inner_crop = Rectangle(top, left, bottom, right)
        new_img_bbox = Rectangle(0, 0, self.new_imsize[0] - 1, self.new_imsize[1] - 1)
        self.inner_crop = new_img_bbox.crop(some_inner_crop)[0]

    @staticmethod
    def _affine_matrix_and_new_canvas_size(imsize, angle_degrees_ccw):
        """
        """
        rows, cols = imsize
        bottom, right = rows - 1, cols - 1
        image_center = (bottom / 2.0, right / 2.0)
        # Opencv uses a left-hand-side coordinate system (col, row), so to get the same effective transform we need to
        # negate the angle.
        affine_matrix = cv2.getRotationMatrix2D(image_center, -angle_degrees_ccw, 1.0)
        source_corners_uniform = np.array([
            [0, 0, 1],
            [0, right, 1],
            [bottom, 0, 1],
            [bottom, right, 1]])
        rotated_corners = affine_matrix.dot(source_corners_uniform.T).T

        # Get the upper and lower calues for rotated coordinates to find out the bounding box.
        rotated_upper_values = np.max(rotated_corners, axis=0)
        rotated_lower_values = np.min(rotated_corners, axis=0)

        # Add extra translation to map the upper left corner of the boundig box to origin (i.e. to get coordinates in
        # the new canvas covering the entire rotated image.
        affine_matrix[:, 2] -= rotated_lower_values
        new_canvas_size = np.ceil(rotated_upper_values - rotated_lower_values + 1).astype(np.int64).tolist()
        return affine_matrix, new_canvas_size

    def __init__(self, imsize: Tuple[int, int], angle_degrees_ccw: int):
        """
        Create rotator for given image size and angle.

        :param imsize: Image shape (height, width).
        :type imsize: Tuple[int, int] or List[int, int]
        :param angle_degrees_ccw: Rotation angle in degrees (counter-clockwise).
        :type angle_degrees_ccw: int

        :Usage Example:

            .. code-block:: python

                rotator = ImageRotator((300, 400), 25)
                rotated = rotator.rotate_img(img, use_inter_nearest=True)
        """
        self.src_imsize = tuple(imsize)
        self.angle_degrees_ccw = angle_degrees_ccw
        # Transform matrix for the RHS (col, row) coordinate system.
        self.affine_matrix, self.new_imsize = self._affine_matrix_and_new_canvas_size(imsize, angle_degrees_ccw)
        # Opencv uses a left-hand-side coordinate system (col, row), so to get the same effective transform we need to
        # negate the angle.
        # Also flip the image size as OpenCV uses cols-rows order.
        self.opencv_affine_matrix, _ = self._affine_matrix_and_new_canvas_size(imsize[::-1], -angle_degrees_ccw)
        self._calc_inner_crop()

    def transform_point(self, point: PointLocation) -> PointLocation:
        """
        Calculates new parameters of PointLocation after rotation.

        :param point: PointLocation to transform.
        :type point: :class:`~supervisely.geometry.point_location.PointLocation`
        :returns: PointLocation after transformation.
        :rtype: :class:`~supervisely.geometry.point_location.PointLocation`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                point = sly.PointLocation(100, 200)
                height, width = 300, 400
                rotator = ImageRotator((height, width), 25)

                rotate_point = rotator.transform_point(point)
                rotate_point_json = rotate_point.to_json()
                print(rotate_point_json)
                # Output:
                # {
                #    "points": {
                #        "exterior": [
                #            [224, 175]
                #        ],
                #        "interior": []
                #    }
                # }
        """
        point_np_uniform = np.array([point.row, point.col, 1])
        transformed_np = self.affine_matrix.dot(point_np_uniform)
        # Unwrap numpy types so that round() produces integer results.
        return PointLocation(row=round(transformed_np[0].item()), col=round(transformed_np[1].item()))

    def rotate_img(self, img: np.ndarray, use_inter_nearest: bool) -> np.ndarray:
        """
        Calculates new parameters of image after rotation.

        :param img: Image to rotate.
        :type img: np.ndarray
        :param use_inter_nearest: If True uses `cv2.INTER_NEAREST <https://docs.opencv.org/3.4/da/d54/group__imgproc__transform.html#gga5bb5a1fea74ea38e1a5445ca803ff121aa5521d8e080972c762467c45f3b70e6c>`_ parameter in rotation, otherwise don't.
        :type use_inter_nearest: bool
        :returns: Rotated image
        :rtype: :class:`np.ndarray`

        :Usage Example:

            .. code-block:: python

                height, width = 300, 400
                rotator = ImageRotator((height, width), 25)
                mask = np.zeros((height, width, 3), dtype=np.uint8)
                rotate_mask = rotator.rotate_img(mask, True)
                print(rotate_mask.shape)
                # Output: (441, 489, 3)
        """
        if use_inter_nearest:
            interp = cv2.INTER_NEAREST  # @TODO: cv2 INTER_NEAREST may shift coords, what to do?
        else:
            interp = cv2.INTER_LANCZOS4
        res = cv2.warpAffine(src=img, M=self.opencv_affine_matrix, dsize=tuple(self.new_imsize[::-1]), flags=interp)
        return res



from supervisely.nn.benchmark.semantic_segmentation.base_vis_metric import (
    SemanticSegmVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import MarkdownWidget


class Acknowledgement(SemanticSegmVisMetric):
    """Acknowledgement section for semantic segmentation reports."""


    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            "acknowledgement",
            "Acknowledgement",
            text=self.vis_texts.markdown_acknowledgement,
        )



from typing import List, Union, Dict

from supervisely.app.widgets import Widget
from supervisely import TagMeta, TagMetaCollection
from supervisely.annotation.tag_meta import TagValueType
from supervisely.app.content import StateJson
from supervisely.imaging.color import rgb2hex


VALUE_TYPE_NAME = {
    str(TagValueType.NONE): "NONE",
    str(TagValueType.ANY_STRING): "TEXT",
    str(TagValueType.ONEOF_STRING): "ONE OF",
    str(TagValueType.ANY_NUMBER): "NUMBER",
}


class TagsListPreview(Widget):
    """Scrollable preview list of tag names; read-only display for TagMetas."""

    def __init__(
        self,
        tag_metas: Union[List[TagMeta], TagMetaCollection] = [],
        max_width: int = 300,
        empty_text: str = None,
        widget_id: int = None,
    ):
        """:param tag_metas: List or TagMetaCollection to display.
        :type tag_metas: Union[List[TagMeta], TagMetaCollection]
        :param max_width: Max width in pixels (min 150).
        :type max_width: int
        :param empty_text: Text when list is empty.
        :type empty_text: str, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: int, optional
        """
        self._tag_metas = tag_metas
        self._max_width = self._get_max_width(max_width)
        self._empty_text = empty_text
        super().__init__(widget_id=widget_id, file_path=__file__)

    def _get_max_width(self, value):
        if value < 150:
            value = 150
        return f"{value}px"

    def get_json_data(self):
        return {
            "maxWidth": self._max_width,
        }

    def get_json_state(self) -> Dict:
        return {
            "tags": [
                {
                    "name": f"<b>{tag_meta.name}</b>",
                    "valueType": VALUE_TYPE_NAME[tag_meta.value_type],
                    "color": rgb2hex(tag_meta.color),
                }
                for tag_meta in self._tag_metas
            ]
        }

    def set(self, tag_metas: Union[List[TagMeta], TagMetaCollection]):
        self._tag_metas = tag_metas
        StateJson()[self.widget_id] = self.get_json_state()
        StateJson().send_changes()



from __future__ import annotations

from typing import Any, Callable, Dict, List, Literal, Optional, Union

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


class Dropdown(Widget):
    """Dropdown menu for selecting an action."""

    class Routes:
        """Route name constants for this widget."""
        VALUE_CHANGED = "value_changed"

    class Item:
        """Single dropdown menu item."""

        def __init__(
            self,
            text: Optional[str] = "",
            disabled: Optional[bool] = False,
            divided: Optional[bool] = False,
            command: Optional[Union[str, int]] = None,
        ) -> Dropdown.Item:
            """
            :param text: Item label.
            :type text: Optional[str]
            :param disabled: Disable item.
            :type disabled: Optional[bool]
            :param divided: Show divider after item.
            :type divided: Optional[bool]
            :param command: Value emitted on click.
            :type command: Optional[Union[str, int]]
            """
            self.text = text
            self.disabled = disabled
            self.divided = divided
            self.command = command

        def to_json(self) -> Dict[str, Union[str, bool, int]]:
            """Return JSON representation of the item.

            Dictionary contains the following fields:
                - text: Item text
                - disabled: If True, item will be disabled
                - divided: If True, item will be divided from the next one
                - command: Item command

            :returns: Dictionary with item data
            :rtype: Dict[str, Union[str, bool, int]]
            """
            return {
                "text": self.text,
                "disabled": self.disabled,
                "divided": self.divided,
                "command": self.command,
            }

    def __init__(
        self,
        items: Optional[List[Dropdown.Item]] = None,
        header: Optional[str] = "Dropdown List",
        trigger: Optional[Literal["hover", "click"]] = "click",
        menu_align: Optional[Literal["start", "end"]] = "end",
        hide_on_click: Optional[bool] = True,
        widget_id: Optional[str] = None,
    ):
        """
        :param items: List of Dropdown.Item.
        :type items: Optional[List[Dropdown.Item]]
        :param header: Header text.
        :type header: Optional[str]
        :param trigger: Dropdown trigger type, one of: hover, click.
        :type trigger: Optional[Literal["hover", "click"]]
        :param menu_align: Dropdown menu alignment, one of: start, end.
        :type menu_align: Optional[Literal["start", "end"]]
        :param hide_on_click: If True, dropdown menu will be hidden after click.
        :type hide_on_click: Optional[bool]
        :param widget_id: Unique widget identifier.
        :type widget_id: Optional[str]

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import Dropdown
                dd = Dropdown(items=[Dropdown.Item("Opt 1", command="a")], header="Options")
        """
        self._items = items
        self._header = header
        self._trigger = trigger
        self._menu_align = menu_align
        self._hide_on_click = hide_on_click
        self._changes_handled = False
        self._clicked_value = None

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _set_items(self):
        return [item.to_json() for item in self._items]

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - trigger: Dropdown trigger type, one of: hover, click
            - items: List of items in the dropdown menu
            - menuAlign: Dropdown menu alignment, one of: start, end
            - hideOnClick: If True, dropdown menu will be hidden after click
            - header: Dropdown header text

        :returns: Dictionary with widget data
        :rtype: Dict[str, Union[str, Any]]
        """
        return {
            "trigger": self._trigger,
            "items": self._set_items(),
            "menuAlign": self._menu_align,
            "hideOnClick": self._hide_on_click,
            "header": self._header,
        }

    def get_json_state(self) -> Dict[str, str]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - clickedValue: current clicked value

        :returns: Dictionary with widget state
        :rtype: Dict[str, str]
        """
        return {"clickedValue": self._clicked_value}

    def get_value(self) -> str:
        """Returns current clicked value.

        :returns: current clicked value
        :rtype: str
        """
        return StateJson()[self.widget_id]["clickedValue"]

    def set_value(self, value: str) -> None:
        """Sets current clicked value.

        :param value: current clicked value
        :type value: str
        """
        self._clicked_value = value
        StateJson()[self.widget_id]["clickedValue"] = self._clicked_value
        StateJson().send_changes()

    def get_items(self) -> List[Dropdown.Item]:
        """Returns list of items in the dropdown menu.

        :returns: list of items in the dropdown menu
        :rtype: List[:class:`~supervisely.app.widgets.dropdown.dropdown.Dropdown.Item`]
        """
        return DataJson()[self.widget_id]["items"]

    def set_items(self, value: List[Dropdown.Item]) -> None:
        """Sets list of items in the dropdown menu.
        This method will overwrite all existing items.
        To add items to the dropdown menu, use :meth:`add_items` instead.

        :param value: list of items in the dropdown menu
        :type value: List[:class:`~supervisely.app.widgets.dropdown.dropdown.Dropdown.Item`]
        """
        if not all(isinstance(item, Dropdown.Item) for item in value):
            raise TypeError("Items must be a list of Dropdown.Item")
        self._items = value
        DataJson()[self.widget_id]["items"] = self._set_items()
        DataJson().send_changes()

    def add_items(self, value: List[Dropdown.Item]) -> None:
        """Adds items to the dropdown menu.
        This method will add items to the existing list.
        To overwrite all existing items, use :meth:`set_items` instead.

        :param value: list of items in the dropdown menu
        :type value: List[:class:`~supervisely.app.widgets.dropdown.dropdown.Dropdown.Item`]
        """
        self._items.extend(value)
        DataJson()[self.widget_id]["items"] = self._set_items()
        DataJson().send_changes()

    def get_header_text(self) -> str:
        """Returns dropdown header text.

        :returns: dropdown header text
        :rtype: str
        """
        return DataJson()[self.widget_id]["header"]

    def set_header_text(self, value: str) -> None:
        """Sets dropdown header text.

        :param value: dropdown header text
        :type value: str
        """
        if type(value) is not str:
            raise TypeError("Header value must be a string")
        self._header = value
        DataJson()[self.widget_id]["header"] = self._header
        DataJson().send_changes()

    def value_changed(self, func: Callable[[str], Any]) -> Callable[[], None]:
        """Decorator for the function to be called when the value is changed.

        :param func: function to be called when the value is changed
        :type func: Callable[[str], Any]
        :returns: decorated function
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(Dropdown.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_value()
            func(res)

        return _click



# coding: utf-8

# docs
from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional

from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection
from supervisely.collection.key_indexed_collection import MultiKeyIndexedCollection


class TagCollection(MultiKeyIndexedCollection):
    """Collection with :class:`~supervisely.annotation.tag.Tag` instances."""

    item_type = Tag

    def __init__(self, items: Optional[List[Tag]] = None):
        """
        :param items: List of :class:`~supervisely.annotation.tag.Tag` instances.
        :type items: list, optional
        :raises :class:`~supervisely.collection.key_indexed_collection.DuplicateKeyError`: if instance with given name already exist

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Create TagMetas (see class TagMeta for additional information about creating TagMetas)
                meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

                seasons = ["Winter", "Spring", "Summer", "Autumn"]
                meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=seasons)

                # Create Tags
                tag_weather = sly.Tag(meta_weather, value="Sunny")
                tag_season = sly.Tag(meta_season, value="Spring")
                tags_arr = [tag_weather, tag_season]

                # Create TagCollection from Tags
                tags = sly.TagCollection(tags_arr)

                # Add item to TagCollection
                meta_potato = sly.TagMeta('potato', sly.TagValueType.NONE)
                tag_potato =sly.Tag(meta_potato)

                # Remember that TagCollection is immutable, and we need to assign new instance of TagCollection to a new variable
                tags = tags.add(tag_potato)

                # You can also add multiple items to collection
                meta_cabbage = sly.TagMeta('cabbage', sly.TagValueType.NONE)
                meta_carrot = sly.TagMeta('carrot', sly.TagValueType.NONE)
                meta_turnip = sly.TagMeta('turnip', sly.TagValueType.NONE)

                tag_cabbage = sly.Tag(meta_cabbage)
                tag_carrot = sly.Tag(meta_carrot)
                tag_turnip = sly.Tag(meta_turnip)

                additional_veggies = [tag_cabbage, tag_carrot, tag_turnip]

                tags = tags.add_items(additional_veggies)

                # Has key, checks if given key exist in collection
                tags.has_key("cabbage")
                # Output: True

                # Find intersection of given list of instances with collection items
                meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
                tag_dog = sly.Tag(meta_dog)

                meta_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
                tag_cat = sly.Tag(meta_cat)

                meta_turtle = sly.TagMeta('turtle', sly.TagValueType.NONE)
                tag_turtle = sly.Tag(meta_turtle)


                tags_animals = sly.TagCollection([tag_dog, tag_cat, tag_turtle])

                tags_intersections = tags.intersection(tags_animals)
                print(tags_intersections.to_json())
                # Output: []

                # Let's add the potato Tag from another collection and compare them again
                tags_animals = tags_animals.add(tag_potato)

                tags_intersections = tags.intersection(tags_animals)
                print(tags_intersections.to_json())
                # Output: [
                #     {
                #         "name":"potato"
                #     }
                # ]

                # Find difference between collection and given list of Tags or TagCollection
                meta_car = sly.TagMeta('car', sly.TagValueType.NONE)
                tag_car = sly.Tag(meta_car)

                meta_bicycle = sly.TagMeta('bicycle', sly.TagValueType.NONE)
                tag_bicycle = sly.Tag(meta_bicycle)

                tags_vehicles = sly.TagCollection([tag_car, tag_bicycle])

                meta_pedestrian = sly.TagMeta('pedestrian', sly.TagValueType.NONE)
                tag_pedestrian = sly.Tag(meta_pedestrian)

                meta_road = sly.TagMeta('road', sly.TagValueType.NONE)
                tag_road = sly.Tag(meta_road)

                difference = tags_vehicles.difference([tag_pedestrian, tag_road])
                print(difference.to_json())
                # Output: [
                #     {
                #         "name":"car"
                #     },
                #     {
                #         "name":"bicycle"
                #     }
                # ]

                # Merge collection and given list of TagMetas
                tags_vehicles = sly.TagMetaCollection([tag_car, tag_bicycle])
                tags_merge = sly.TagMetaCollection([tag_pedestrian, tag_road])

                merged_collections = tags_vehicles.merge(tags_merge)
                print(merged_collections.to_json())
                # Output: [
                #     {
                #         "name":"pedestrian"
                #     },
                #     {
                #         "name":"road"
                #     },
                #     {
                #         "name":"car"
                #     },
                #     {
                #         "name":"bicycle"
                #     }
                # ]
        """
        super().__init__(items)

    def to_json(self) -> List[Dict]:
        """
        Convert the TagCollection to a list of json dicts. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: List of dicts in json format
        :rtype: :class:`List[dict]`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

                seasons = ["Winter", "Spring", "Summer", "Autumn"]
                meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=seasons)

                tag_weather = sly.Tag(meta_weather, value="Sunny")
                tag_season = sly.Tag(meta_season, value="Spring")

                tags_arr = [tag_weather, tag_season]
                tags = sly.TagCollection(tags_arr)

                tags_json = tags.to_json()
                # Output: [
                #     {
                #         "name":"Weather",
                #         "value":"Sunny"
                #     },
                #     {
                #         "name":"Season",
                #         "value":"Spring"
                #     }
                # ]
        """
        return [tag.to_json() for tag in self]

    @classmethod
    def from_json(cls, data: List[Dict], tag_meta_collection: TagMetaCollection) -> TagCollection:
        """
        Convert a list with dicts in json format to TagCollection. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[dict]
        :param tag_meta_collection: TagMetaCollection object.
        :type tag_meta_collection: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :returns: TagCollection object
        :rtype: :class:`~supervisely.annotation.tag_collection.TagCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Initialize TagMetaCollection

                meta_weather = sly.TagMeta('Weather', sly.TagValueType.ANY_STRING)

                seasons = ["Winter", "Spring", "Summer", "Autumn"]
                meta_season = sly.TagMeta('Season', sly.TagValueType.ONEOF_STRING, possible_values=seasons)

                metas_arr = [meta_weather, meta_season]
                tag_metas = sly.TagMetaCollection(metas_arr)

                data = [
                {
                    "name":"Weather",
                    "value":"Sunny"
                },
                {
                    "name":"Season",
                    "value":"Spring"
                }
                ]

                tags = sly.TagCollection.from_json(data, tag_metas)
        """
        tags = [cls.item_type.from_json(tag_json, tag_meta_collection) for tag_json in data]
        return cls(tags)

    def __str__(self):
        return "Tags:\n" + super(TagCollection, self).__str__()

    @classmethod
    def from_api_response(
        cls,
        data: List[Dict],
        tag_meta_collection: TagMetaCollection,
        id_to_tagmeta: Optional[Dict[int, TagMeta]] = None,
    ) -> TagCollection:
        """
        Create a TagCollection object from API response data.

        :param data: API response data.
        :type data: List[Dict]
        :param tag_meta_collection: TagMetaCollection object.
        :type tag_meta_collection: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :param id_to_tagmeta: Mapping of tag IDs to tag metadata.
        :type id_to_tagmeta: Optional[Dict[int, :class:`~supervisely.annotation.tag_meta.TagMeta`]]
        :returns: TagCollection object.
        :rtype: :class:`~supervisely.annotation.tag_collection.TagCollection`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 17200
                image_id = 19369643
                project_meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(project_meta_json)
                image_info = api.image.get_info_by_id(image_id)

                tag_collection = sly.TagCollection.from_api_response(
                    image_info.tags, project_meta.tag_metas
                )
                print(tag_collection)

                # Output:
                # Tags:
                # +-------+--------------+-------+
                # |  Name |  Value type  | Value |
                # +-------+--------------+-------+
                # | Lemon | oneof_string |  big  |
                # |  Kiwi |     none     |  None |
                # +-------+--------------+-------+
        """

        if id_to_tagmeta is None:
            id_to_tagmeta = tag_meta_collection.get_id_mapping()
        tags = []
        for tag_json in data:
            tag_meta_id = tag_json["tagId"]
            tag_meta = id_to_tagmeta[tag_meta_id]
            tag_json["name"] = tag_meta.name
            tag = cls.item_type.from_json(tag_json, tag_meta_collection)
            tags.append(tag)
        return cls(tags)

    def __iter__(self) -> Iterator[Tag]:
        return next(self)



import os
import tempfile
from collections import defaultdict, namedtuple
from typing import Dict, List, Optional

import cv2
import nrrd
import numpy as np

from supervisely import ProjectMeta, is_development, logger
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.tag import Tag
from supervisely.api.api import Api, ApiContext
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.convert.image.image_helper import read_tiff_image
from supervisely.imaging.image import is_valid_ext
from supervisely.io.fs import dirs_filter, get_file_ext, list_files
from supervisely.project.project_settings import LabelingInterface
from supervisely.project.project_type import _MULTISPECTRAL_TAG_NAME

SPLIT_TO_CHANNELS_DIR_NAME = "split"
UPLOAD_AS_IMAGES_DIR_NAME = "images"
ImageGroup = namedtuple("ImageGroup", ["split", "upload"])


class MultiSpectralImageConverter(ImageConverter):
    """Imports multispectral images (split channels or composite) for multispectral labeling interface."""

    def __str__(self):
        return AvailableImageConverters.MULTISPECTRAL

    def validate_labeling_interface(self) -> bool:
        """Only multispectral labeling interface can be used for multispectral images."""
        return self._labeling_interface == LabelingInterface.MULTISPECTRAL

    def validate_format(self) -> bool:
        logger.debug(f"Validating format: {self.__str__()}")
        group_map = self._find_image_directories()
        if not group_map:
            logger.debug(f"No multispectral images found in {self._input_data}.")
            return False
        else:
            self._group_map = group_map
            logger.debug(f"Found multispectral images in {self._input_data}.")
            return True

    def _find_image_directories(self) -> Dict[str, ImageGroup]:
        group_map = defaultdict(ImageGroup)
        for checked_directory in dirs_filter(self._input_data, self._check_directory):
            split_images = []
            upload_images = []

            split_dir = os.path.join(checked_directory, SPLIT_TO_CHANNELS_DIR_NAME)

            images_dir = os.path.join(checked_directory, UPLOAD_AS_IMAGES_DIR_NAME)

            if os.path.isdir(split_dir):
                split_images = list_files(split_dir)
            if os.path.isdir(images_dir):
                upload_images = list_files(images_dir)

            group_map[checked_directory] = ImageGroup(split_images, upload_images)

        return group_map

    def _check_directory(self, path: str) -> bool:
        split_dir = os.path.join(path, SPLIT_TO_CHANNELS_DIR_NAME)
        images_dir = os.path.join(path, UPLOAD_AS_IMAGES_DIR_NAME)

        if os.path.isdir(split_dir) or os.path.isdir(images_dir):
            return True
        return False

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 50,
        log_progress=True,
    ) -> None:
        """Upload converted data to Supervisely"""
        dataset = api.dataset.get_info_by_id(dataset_id)
        project_id = dataset.project_id
        api.project.set_multispectral_settings(project_id)

        meta_json = api.project.get_meta(project_id, with_settings=True)
        meta = ProjectMeta.from_json(meta_json)
        group_tag_meta = meta.get_tag_meta(_MULTISPECTRAL_TAG_NAME)

        groups_count = len(self._group_map)
        if log_progress:
            progress, progress_cb = self.get_progress(groups_count, "Processing groups...")
        else:
            progress_cb = None

        for group_path, image_group in self._group_map.items():
            logger.info(f"Found files in {group_path}.")

            images, channels = [], []
            nrrd_paths, nrrds_anns, nrrd_names = [], [], []

            images_to_split, images_to_upload = image_group.split, image_group.upload

            group_name = os.path.basename(group_path)
            for image_to_upload in images_to_upload:
                images.append(os.path.join(group_path, image_to_upload))
            for image_to_split in images_to_split:
                image = self._get_image(os.path.join(group_path, image_to_split))
                if image is None:
                    continue

                # image has wide range of pixel values, we will upload it as nrrd
                # (supports `windowing` feature in the Image Labeling Toolbox)
                img_channels = self._get_image_channels(image)
                for idx, channel in enumerate(img_channels):
                    if channel.dtype != np.uint8 or np.max(channel) > 255:
                        nrrd_path = self._prepare_nrrd(channel, image_to_split, idx)
                        nrrd_paths.append(nrrd_path)
                        nrrd_names.append(os.path.basename(nrrd_path))
                        group_tag = Tag(meta=group_tag_meta, value=group_name)
                        nrrds_anns.append(Annotation(channel.shape).add_tag(group_tag))
                    else:
                        channels.append(channel)

            with ApiContext(
                api=api, project_id=project_id, dataset_id=dataset_id, project_meta=meta
            ):
                img_count = len(channels) + len(images) + len(nrrd_paths)
                if log_progress:
                    _, _progress_cb = self.get_progress(img_count, f"Uploading {group_name}...")
                else:
                    _progress_cb = None

                if channels or images:
                    api.image.upload_multispectral(
                        dataset.id, group_name, channels, images, _progress_cb
                    )
                if nrrd_paths:
                    img_infos = api.image.upload_paths(
                        dataset.id,
                        nrrd_names,
                        nrrd_paths,
                        _progress_cb,
                        conflict_resolution="rename",
                    )
                    image_ids = [image_info.id for image_info in img_infos]
                    api.annotation.upload_anns(image_ids, nrrds_anns)

            if progress_cb is not None:
                progress_cb(1)

        if log_progress:
            if is_development():
                progress.close()
        logger.info(f"Dataset '{dataset.name}' has been successfully uploaded.")

    def _get_image(self, file_path: str) -> Optional[np.ndarray]:
        file_ext = get_file_ext(file_path).lower()
        logger.debug(f"Working with file {file_path} with extension {file_ext}.")

        if file_ext == ".nrrd":
            logger.debug(f"Found nrrd file: {file_path}.")
            image, _ = nrrd.read(file_path)
        elif file_ext in [".tif", ".tiff"]:
            image = read_tiff_image(file_path)
        elif is_valid_ext(file_ext):
            logger.debug(f"Found image file: {file_path}.")
            image = cv2.imread(file_path)
        else:
            logger.warning(f"File {file_path} has unsupported extension.")
            return
        if image is None:
            logger.warning(f"Failed to read image {file_path}.")
            return

        return image

    def _get_image_channels(self, image) -> List[np.ndarray]:
        if image is None:
            return []

        # Handle single-channel images (grayscale)
        if len(image.shape) == 2:
            logger.debug(f"Image has single channel.")
            return [image]

        # Handle multi-channel images
        return [image[:, :, i] for i in range(image.shape[2])]

    def _prepare_nrrd(self, image: np.ndarray, image_name: str, channel_idx: int) -> str:
        header = {"sizes": image.shape, "type": str(image.dtype), "dimension": len(image.shape)}

        _dir = os.path.join(tempfile.gettempdir(), "supervisely_multispectral_nrrds")
        os.makedirs(_dir, exist_ok=True)

        base_name = os.path.basename(image_name)
        name = f"{os.path.splitext(base_name)[0]}_{channel_idx}.nrrd"
        path = os.path.join(_dir, name)

        nrrd.write(path, image, header)
        return path



from typing import Dict, Optional, Union

from supervisely.app.widgets import Editor, Input, Text, TextArea, Widget


class CopyToClipboard(Widget):
    """Wraps Editor, Text, TextArea, or Input with a copy-to-clipboard button."""

    def __init__(
        self,
        content: Optional[Union[Editor, Text, TextArea, Input, str]] = "",
        widget_id: Optional[str] = None,
    ):
        """
        :param content: Editor, Text, TextArea, Input, or str.
        :type content: Optional[Union[Editor, Text, TextArea, Input, str]]
        :param widget_id: Unique widget identifier.
        :type widget_id: Optional[str]
        :raises TypeError: If content type not supported.

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import CopyToClipboard, Text
                w = CopyToClipboard(Text("Copy me"))
        """
        self._content = content

        if not isinstance(content, (str, Editor, Text, TextArea, Input)):
            raise TypeError(
                f"Supported types: str, Editor, Text, TextArea, Input. Your type: {type(content).__name__}"
            )
        if isinstance(content, str):
            self._content_widget_type = "str"
            self._curr_prop_name = None
            self._content_value = content
        else:
            if isinstance(content, (Editor, Input)):
                self._content_widget_type = "input"
                self._curr_prop_name = "value" if isinstance(content, Input) else "text"
            elif isinstance(content, (Text, TextArea)):
                self._content_widget_type = "text"
                self._curr_prop_name = "value" if isinstance(content, TextArea) else "text"
            self._content_value = content.get_value()

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Union[str, Dict]]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - content: content to be copied
            - curr_property: current property of the content widget

        :returns: dictionary with widget data
        :rtype: Dict[str, Union[str, Dict]]
        """
        return {"content": self._content_value, "curr_property": self._curr_prop_name}

    def get_json_state(self) -> Dict[str, Union[str, Dict]]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - content: content to be copied
            - curr_property: current property of the content widget

        :returns: dictionary with widget state
        :rtype: Dict[str, Union[str, Dict]]
        """
        return {"content": self._content_value, "curr_property": self._curr_prop_name}

    def get_content(self) -> Union[Editor, Input, Text, TextArea, str]:
        """Returns content of the widget.

        :returns: content of the widget
        :rtype: Union[:class:`~supervisely.app.widgets.editor.editor.Editor`, :class:`~supervisely.app.widgets.input.input.Input`, :class:`~supervisely.app.widgets.text.text.Text`, :class:`~supervisely.app.widgets.textarea.textarea.TextArea`, str]
        """
        return self._content



import os

from supervisely._utils import abs_url, is_debug_with_sly_net, is_development
from supervisely.api.file_api import FileInfo
from supervisely.app import DataJson
from supervisely.app.widgets import Widget


class FolderThumbnail(Widget):
    """Compact card showing a Team Files folder with path and link."""

    def __init__(self, info=None, widget_id: str = None):
        """
        :param info: FileInfo from Team Files (folder). Can be set later with set().
        :type info: FileInfo, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._id: int = None
        self._info: info = None
        self._description: str = None
        self._url: str = None
        self._set_info(info)

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "id": self._id,
            "name": "Folder in Team files",
            "description": self._description,
            "url": self._url,
            "description": self._description,
            "icon": {
                "className": "zmdi zmdi-folder-person",
                "color": "#4977ff",
                "bgColor": "#ddf2ff",
            },
        }

    def get_json_state(self):
        return None

    def _set_info(self, info: FileInfo = None):
        if info is None:
            return
        self._id = info.id
        self._info = info
        self._description = os.path.dirname(info.path)
        if is_development() or is_debug_with_sly_net():
            self._url = abs_url(f"files/{info.id}")
        else:
            self._url = f"/files/{info.id}"

    def set(self, info):
        self._set_info(info)
        self.update_data()
        DataJson().send_changes()



import json
import os
import threading
import time
from typing import Any, Dict, List, Optional, Union

from cacheout import Cache
from cachetools import LRUCache
from fastapi import Form, Request, Response, UploadFile, status
from fastapi.responses import JSONResponse

from supervisely import Label, ObjClass, ProjectMeta
from supervisely import env as sly_env
from supervisely._utils import rand_str
from supervisely.app.content import get_data_dir
from supervisely.geometry.bitmap import Bitmap
from supervisely.imaging import image as sly_image
from supervisely.io.fs import silent_remove
from supervisely.nn.inference import Inference
from supervisely.nn.inference.interactive_segmentation import functional
from supervisely.nn.prediction_dto import PredictionSegmentation
from supervisely.sly_logger import logger

try:
    from typing import Literal
except ImportError:
    # for compatibility with python 3.7
    from typing_extensions import Literal


class InteractiveSegmentation(Inference):
    """Inference server base for click-based interactive segmentation (smart tool) backends."""

    class Click:
        """Single user click point (x/y) with positive/negative flag."""

        def __init__(self, x, y, is_positive):
            """
            :param x: X coordinate.
            :type x: int
            :param y: Y coordinate.
            :type y: int
            :param is_positive: True for foreground.
            :type is_positive: bool
            """
            self.x = x
            self.y = y
            self.is_positive = is_positive

        def __repr__(self) -> str:
            return f"{self.__class__.__name__}: {str(self.__dict__)}"

    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[Union[Dict[str, Any], str]] = None,
        sliding_window_mode: Optional[Literal["basic", "advanced", "none"]] = "basic",
        use_gui: Optional[bool] = False,
    ):
        """See :class:`~supervisely.nn.inference.inference.Inference` for params."""
        _smart_cache_ttl = sly_env.smart_cache_ttl()
        _fast_cache_ttl = max(1, _smart_cache_ttl // 2)
        Inference.__init__(self, model_dir, custom_inference_settings, sliding_window_mode, use_gui)
        self._class_names = ["mask_prediction"]
        color = [255, 0, 0]
        self._model_meta = ProjectMeta([ObjClass(self._class_names[0], Bitmap, color)])
        self._inference_image_lock = threading.Lock()
        self._inference_image_cache = Cache(ttl=_fast_cache_ttl)
        self._init_mask_cache = LRUCache(maxsize=100)  # cache of sly.Bitmaps

        if not self._use_gui:
            try:
                self.load_on_device(model_dir, "cuda")
            except RuntimeError:
                self.load_on_device(model_dir, "cpu")
                logger.warn("Failed to load model on CUDA device.")

        logger.debug(
            "Smart cache params",
            extra={"ttl": _smart_cache_ttl, "maxsize": sly_env.smart_cache_size()},
        )

    def get_info(self) -> dict:
        info = super().get_info()
        info["task type"] = "interactive segmentation"
        info["videos_support"] = False
        info["async_video_inference_support"] = False
        info["tracking_on_videos_support"] = False
        return info

    def _get_obj_class_shape(self):
        return Bitmap

    def _create_label(self, dto: PredictionSegmentation):
        classes = self.get_classes()
        assert len(classes) == 1, "InteractiveSegmentation can't be used for multi-class inference"
        obj_class = self.model_meta.get_obj_class(classes[0])
        if not dto.mask.any():  # skip empty masks
            logger.debug(f"Mask of class {dto.class_name} is empty and will be skipped")
            return None
        geometry = Bitmap(dto.mask, extra_validation=False)
        label = Label(geometry, obj_class)
        return label

    def predict(
        self,
        image_path: str,
        clicks: List[Click],
        settings: Dict[str, Any],
    ) -> PredictionSegmentation:
        raise NotImplementedError("Have to be implemented in child class")

    def get_classes(self) -> List[str]:
        return self._class_names

    def serve(self):
        super().serve()
        server = self._app.get_server()
        self.cache.add_cache_endpoint(server)

        @server.post("/smart_segmentation")
        def smart_segmentation(response: Response, request: Request):
            logger.debug(
                f"smart_segmentation inference: context=",
                extra={**request.state.context, "api_token": "***"},
            )

            # Parse request
            try:
                state = request.state.state
                settings = self._get_inference_settings(state)
                smtool_state = request.state.context
                api = request.state.api
                crop = smtool_state["crop"]
                positive_clicks, negative_clicks = (
                    smtool_state["positive"],
                    smtool_state["negative"],
                )
                if len(positive_clicks) + len(negative_clicks) == 0:
                    logger.warn("No clicks received.")
                    response = {
                        "origin": None,
                        "bitmap": None,
                        "success": True,
                        "error": None,
                    }
                    return response
            except Exception as exc:
                logger.warn("Error parsing request:" + str(exc), exc_info=True)
                response.status_code = status.HTTP_400_BAD_REQUEST
                return {"message": "400: Bad request.", "success": False}

            # Pre-process clicks
            clicks = [{**click, "is_positive": True} for click in positive_clicks]
            clicks += [{**click, "is_positive": False} for click in negative_clicks]
            clicks = functional.transform_clicks_to_crop(crop, clicks)
            is_in_bbox = functional.validate_click_bounds(crop, clicks)
            if not is_in_bbox:
                logger.warn(f"Invalid value: click is out of bbox bounds.")
                return {
                    "origin": None,
                    "bitmap": None,
                    "success": True,
                    "error": None,
                }

            # Download the image if is not in Cache
            app_dir = get_data_dir()
            hash_str = functional.get_hash_from_context(smtool_state)
            if hash_str not in self._inference_image_cache:
                logger.debug(f"downloading image: {hash_str}")
                image_np = functional.download_image_from_context(
                    smtool_state,
                    api,
                    app_dir,
                    self.cache.download_image,
                    self.cache.download_frame,
                    self.cache.download_image_by_hash,
                )
                self._inference_image_cache.set(hash_str, image_np)
            else:
                logger.debug(f"image found in cache: {hash_str}")
                image_np = self._inference_image_cache.get(hash_str)

            # Crop the image
            image_np = functional.crop_image(crop, image_np)
            image_path = os.path.join(app_dir, f"{time.time()}_{rand_str(10)}.jpg")
            sly_image.write(image_path, image_np)

            # Prepare init_mask (only for images)
            figure_id = smtool_state.get("figure_id")
            image_id = smtool_state.get("image_id")
            if smtool_state.get("init_figure") is True and image_id is not None:
                # Download and save in Cache
                init_mask = functional.download_init_mask(api, figure_id, image_id)
                self._init_mask_cache[figure_id] = init_mask
            elif self._init_mask_cache.get(figure_id) is not None:
                # Load from Cache
                init_mask = self._init_mask_cache[figure_id]
            else:
                init_mask = None
            if init_mask is not None:
                img_info = api.image.get_info_by_id(image_id)
                h, w = img_info.height, img_info.width
                init_mask = functional.bitmap_to_mask(init_mask, h, w)
                init_mask = functional.crop_image(crop, init_mask)
                assert init_mask.shape[:2] == image_np.shape[:2]
            settings["init_mask"] = init_mask

            # Predict
            self._inference_image_lock.acquire()
            try:
                logger.debug(f"predict: {smtool_state['request_uid']}")
                clicks_to_predict = [self.Click(c["x"], c["y"], c["is_positive"]) for c in clicks]
                pred_mask = self.predict(image_path, clicks_to_predict, settings).mask
            finally:
                logger.debug(f"predict done: {smtool_state['request_uid']}")
                self._inference_image_lock.release()
                silent_remove(image_path)

            if pred_mask.any():
                bitmap = Bitmap(pred_mask)
                bitmap_origin, bitmap_data = functional.format_bitmap(bitmap, crop)
                logger.debug(f"smart_segmentation inference done!")
                response = {
                    "origin": bitmap_origin,
                    "bitmap": bitmap_data,
                    "success": True,
                    "error": None,
                }
            else:
                logger.debug(f"Predicted mask is empty.")
                response = {
                    "origin": None,
                    "bitmap": None,
                    "success": True,
                    "error": None,
                }
            return response

        @server.post("/smart_segmentation_batch")
        def smart_segmentation_batch(response: Response, request: Request):
            result = []
            logger.debug(
                f"smart_segmentation inference: context=",
                extra={**request.state.context, "api_token": "***"},
            )

            # Parse request
            try:
                state = request.state.state
                settings = self._get_inference_settings(state)
                api = request.state.api
                smtool_states = request.state.context.get("states", [])
            except Exception as exc:
                logger.warn("Error parsing request:" + str(exc), exc_info=True)
                response.status_code = status.HTTP_400_BAD_REQUEST
                return {"message": "400: Bad request.", "success": False}

            for smtool_state in smtool_states:
                crop = smtool_state["crop"]
                positive_clicks, negative_clicks = (
                    smtool_state["positive"],
                    smtool_state["negative"],
                )
                if len(positive_clicks) + len(negative_clicks) == 0:
                    logger.warn("No clicks received.")
                    result.append(
                        {
                            "origin": None,
                            "bitmap": None,
                            "success": True,
                            "error": None,
                        }
                    )
                    continue

                # Pre-process clicks
                clicks = [{**click, "is_positive": True} for click in positive_clicks]
                clicks += [{**click, "is_positive": False} for click in negative_clicks]
                clicks = functional.transform_clicks_to_crop(crop, clicks)
                is_in_bbox = functional.validate_click_bounds(crop, clicks)
                if not is_in_bbox:
                    logger.warn(f"Invalid value: click is out of bbox bounds.")
                    result.append(
                        {
                            "origin": None,
                            "bitmap": None,
                            "success": True,
                            "error": None,
                        }
                    )
                    continue

                # Download the image if is not in Cache
                app_dir = get_data_dir()
                hash_str = functional.get_hash_from_context(smtool_state)
                if hash_str not in self._inference_image_cache:
                    logger.debug(f"downloading image: {hash_str}")
                    image_np = functional.download_image_from_context(
                        smtool_state,
                        api,
                        app_dir,
                        self.cache.download_image,
                        self.cache.download_frame,
                        self.cache.download_image_by_hash,
                    )
                    self._inference_image_cache.set(hash_str, image_np)
                else:
                    logger.debug(f"image found in cache: {hash_str}")
                    image_np = self._inference_image_cache.get(hash_str)

                # Crop the image
                image_np = functional.crop_image(crop, image_np)
                image_path = os.path.join(app_dir, f"{time.time()}_{rand_str(10)}.jpg")
                sly_image.write(image_path, image_np)

                # Prepare init_mask (only for images)
                figure_id = smtool_state.get("figure_id")
                image_id = smtool_state.get("image_id")
                if smtool_state.get("init_figure") is True and image_id is not None:
                    # Download and save in Cache
                    init_mask = functional.download_init_mask(api, figure_id, image_id)
                    self._init_mask_cache[figure_id] = init_mask
                elif self._init_mask_cache.get(figure_id) is not None:
                    # Load from Cache
                    init_mask = self._init_mask_cache[figure_id]
                else:
                    init_mask = None
                if init_mask is not None:
                    img_info = api.image.get_info_by_id(image_id)
                    h, w = img_info.height, img_info.width
                    init_mask = functional.bitmap_to_mask(init_mask, h, w)
                    init_mask = functional.crop_image(crop, init_mask)
                    assert init_mask.shape[:2] == image_np.shape[:2]
                settings["init_mask"] = init_mask

                # Predict
                self._inference_image_lock.acquire()
                try:
                    logger.debug(f"predict: {smtool_state['request_uid']}")
                    clicks_to_predict = [
                        self.Click(c["x"], c["y"], c["is_positive"]) for c in clicks
                    ]
                    pred_mask = self.predict(image_path, clicks_to_predict, settings).mask
                finally:
                    logger.debug(f"predict done: {smtool_state['request_uid']}")
                    self._inference_image_lock.release()
                    silent_remove(image_path)

                if pred_mask.any():
                    bitmap = Bitmap(pred_mask)
                    bitmap_origin, bitmap_data = functional.format_bitmap(bitmap, crop)
                    logger.debug(f"smart_segmentation inference done!")
                    result.append(
                        {
                            "origin": bitmap_origin,
                            "bitmap": bitmap_data,
                            "success": True,
                            "error": None,
                        }
                    )
                else:
                    logger.debug(f"Predicted mask is empty.")
                    result.append(
                        {
                            "origin": None,
                            "bitmap": None,
                            "success": True,
                            "error": None,
                        }
                    )
            return result

        @server.post("/smart_segmentation_files")
        def smart_segmentation_files(
            request: Request, files: List[UploadFile], settings: str = Form("{}")
        ):
            result = []
            settings = json.loads(settings)
            logger.debug(
                f"smart_segmentation inference: context=",
                extra={settings},
            )
            smtool_states = settings.get("state", [])
            inf_settings = self._get_inference_settings(settings)
            for file, smtool_state in zip(files, smtool_states):
                # Parse request
                try:
                    crop = smtool_state["crop"]
                    positive_clicks, negative_clicks = (
                        smtool_state["positive"],
                        smtool_state["negative"],
                    )
                    if len(positive_clicks) + len(negative_clicks) == 0:
                        logger.warn("No clicks received.")
                        result.append(
                            {
                                "origin": None,
                                "bitmap": None,
                                "success": True,
                                "error": None,
                            }
                        )
                        continue
                except Exception as exc:
                    logger.warn("Error parsing request:" + str(exc), exc_info=True)
                    return JSONResponse(
                        {"message": "400: Bad request.", "success": False}, status_code=400
                    )

                # Pre-process clicks
                clicks = [{**click, "is_positive": True} for click in positive_clicks]
                clicks += [{**click, "is_positive": False} for click in negative_clicks]
                clicks = functional.transform_clicks_to_crop(crop, clicks)
                is_in_bbox = functional.validate_click_bounds(crop, clicks)
                if not is_in_bbox:
                    logger.warn(f"Invalid value: click is out of bbox bounds.")
                    result.append(
                        {
                            "origin": None,
                            "bitmap": None,
                            "success": True,
                            "error": None,
                        }
                    )
                    continue

                # Crop the image
                app_dir = get_data_dir()
                image_np = sly_image.read_bytes(file.file.read())
                hash_str = functional.get_hash_from_context(smtool_state)
                self._inference_image_cache.set(hash_str, image_np)
                image_np = functional.crop_image(crop, image_np)
                image_path = os.path.join(app_dir, f"{time.time()}_{rand_str(10)}.jpg")
                sly_image.write(image_path, image_np)

                # Prepare init_mask (only for images)
                geom_data = smtool_state.get("geometry")
                init_mask = Bitmap.from_json(geom_data) if geom_data is not None else None
                if init_mask is not None:
                    h, w = image_np.shape[:2]
                    init_mask = functional.bitmap_to_mask(init_mask, h, w)
                    init_mask = functional.crop_image(crop, init_mask)
                    assert init_mask.shape[:2] == image_np.shape[:2]
                inf_settings["init_mask"] = init_mask

                # Predict
                self._inference_image_lock.acquire()
                try:
                    logger.debug(f"predict: {smtool_state['request_uid']}")
                    clicks_to_predict = [
                        self.Click(c["x"], c["y"], c["is_positive"]) for c in clicks
                    ]
                    pred_mask = self.predict(image_path, clicks_to_predict, settings).mask
                finally:
                    logger.debug(f"predict done: {smtool_state['request_uid']}")
                    self._inference_image_lock.release()
                    silent_remove(image_path)

                if pred_mask.any():
                    bitmap = Bitmap(pred_mask)
                    bitmap_origin, bitmap_data = functional.format_bitmap(bitmap, crop)
                    logger.debug(f"smart_segmentation inference done!")
                    result.append(
                        {
                            "origin": bitmap_origin,
                            "bitmap": bitmap_data,
                            "success": True,
                            "error": None,
                        }
                    )
                else:
                    logger.debug(f"Predicted mask is empty.")
                    result.append(
                        {
                            "origin": None,
                            "bitmap": None,
                            "success": True,
                            "error": None,
                        }
                    )
            return result



from dataclasses import dataclass, field
from datetime import datetime
from os import path as osp
from pathlib import Path
from typing import Dict, Generator, List, Tuple

import numpy as np

from supervisely import fs, logger
from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

DIR_NAMES = [
    "CAM_BACK",
    "CAM_BACK_LEFT",
    "CAM_BACK_RIGHT",
    "CAM_FRONT",
    "CAM_FRONT_LEFT",
    "CAM_FRONT_RIGHT",
    "LIDAR_TOP",
    # "RADAR_FRONT",
    # "RADAR_FRONT_LEFT",
    # "RADAR_FRONT_RIGHT",
    # "RADAR_BACK_LEFT",
    # "RADAR_BACK_RIGHT",
]

TABLE_NAMES = [
    "category",
    "attribute",
    "visibility",
    "instance",
    "sensor",
    "calibrated_sensor",
    "ego_pose",
    "log",
    "scene",
    "sample",
    "sample_data",
    "sample_annotation",
    "map",
]


def trim_description(description: str, max_length: int = 255) -> str:
    if len(description) > max_length:
        sentences = description.split(".")
        trimmed_description = ""
        for sentence in sentences:
            if len(trimmed_description) + len(sentence) + 1 > max_length:
                break
            trimmed_description += sentence + "."
        description = trimmed_description.strip()
    return description


@dataclass
class AnnotationObject:
    """
    A class to represent an annotation object in the NuScenes dataset.

    :param name: The name of the annotation object
    :type name: str
    :param bbox: The bounding box coordinates in NuScenes format
    :type bbox: np.ndarray
    :param token: The unique token identifying the annotation object
    :type token: str
    :param instance_token: The instance token associated with the annotation object
    :type instance_token: str
    :param parent_token: The token of instance preceding the current object instance
    :type parent_token: str
    :param category: The class name of the annotation object
    :type category: str
    :param attributes: The attribute names associated with the annotation object
    :type attributes: List[str]
    :param visibility: The visibility level of the annotation object
    :type visibility: str
    """
    name: str
    bbox: np.ndarray
    token: str
    instance_token: str
    parent_token: str
    category: str
    attributes: List[str]
    visibility: str

    def to_supervisely(self) -> Cuboid3d:
        box = self.convert_nuscenes_to_BEVBox3D()

        bbox = box.to_xyzwhlr()
        dim = bbox[[3, 5, 4]]
        pos = bbox[:3] + [0, 0, dim[1] / 2]
        yaw = bbox[-1]

        position = Vector3d(float(pos[0]), float(pos[1]), float(pos[2]))
        rotation = Vector3d(0, 0, float(-yaw))
        dimension = Vector3d(float(dim[0]), float(dim[2]), float(dim[1]))
        geometry = Cuboid3d(position, rotation, dimension)

        return geometry

    def convert_nuscenes_to_BEVBox3D(self):
        import open3d as o3d  # pylint: disable=import-error

        box = self.bbox
        center = [float(box[0]), float(box[1]), float(box[2])]
        size = [float(box[3]), float(box[5]), float(box[4])]
        ry = float(box[6])
        yaw = ry - np.pi
        yaw = yaw - np.floor(yaw / (2 * np.pi) + 0.5) * 2 * np.pi
        world_cam = None
        return o3d.ml.datasets.utils.BEVBox3D(center, size, yaw, self.name, -1.0, world_cam)


class CamData:
    """Handles camera sensor data from nuScenes, including lidar-to-camera transformations."""

    def __init__(
        self, nuscenes, sensor_name: str, sensor_token: str, cs_record: dict, ego_record: dict
    ):
        """:param nuscenes: The nuScenes dataset instance.
        :type nuscenes: NuScenes
        :param sensor_name: The name of the camera sensor.
        :type sensor_name: str
        :param sensor_token: The token identifying the specific sensor sample.
        :type sensor_token: str
        :param cs_record: The calibrated sensor record for the lidar.
        :type cs_record: dict
        :param ego_record: The ego pose record for the lidar.
        :type ego_record: dict
        """
        from nuscenes import NuScenes  # pylint: disable=import-error
        from nuscenes.utils.data_classes import (  # pylint: disable=import-error
            transform_matrix,
        )
        from pyquaternion import Quaternion  # pylint: disable=import-error

        nuscenes: NuScenes = nuscenes

        img_path, _, _ = nuscenes.get_sample_data(sensor_token)
        if not osp.exists(img_path):
            return None

        sd_record_cam = nuscenes.get("sample_data", sensor_token)
        cs_record_cam = nuscenes.get("calibrated_sensor", sd_record_cam["calibrated_sensor_token"])
        ego_record_cam = nuscenes.get("ego_pose", sd_record_cam["ego_pose_token"])
        lid_to_ego = transform_matrix(
            cs_record["translation"],
            Quaternion(cs_record["rotation"]),
            inverse=False,
        )
        lid_ego_to_world = transform_matrix(
            ego_record["translation"],
            Quaternion(ego_record["rotation"]),
            inverse=False,
        )
        world_to_cam_ego = transform_matrix(
            ego_record_cam["translation"],
            Quaternion(ego_record_cam["rotation"]),
            inverse=True,
        )
        ego_to_cam = transform_matrix(
            cs_record_cam["translation"],
            Quaternion(cs_record_cam["rotation"]),
            inverse=True,
        )
        velo_to_cam = np.dot(
            ego_to_cam, np.dot(world_to_cam_ego, np.dot(lid_ego_to_world, lid_to_ego))
        )
        velo_to_cam_rot = velo_to_cam[:3, :3]
        velo_to_cam_trans = velo_to_cam[:3, 3]

        self.name = sensor_name
        self.path = str(img_path)
        self.imsize = (sd_record_cam["width"], sd_record_cam["height"])
        self.extrinsic = np.hstack((velo_to_cam_rot, velo_to_cam_trans.reshape(3, 1)))
        self.intrinsic = np.asarray(cs_record_cam["camera_intrinsic"])

    def get_info(self, timestamp: str) -> Tuple[str, Dict]:
        """
        Generates image info based on the camera data.

        :param timestamp: The timestamp associated with the image
        :type timestamp: str
        :returns: A tuple containing the image path and a dictionary with image metadata.
        :rtype: tuple
        """
        sensors_to_skip = ["_intrinsic", "_extrinsic", "_imsize"]
        if not any([self.name.endswith(s) for s in sensors_to_skip]):
            image_name = fs.get_file_name_with_ext(self.path)
            sly_path_img = osp.join(osp.dirname(self.path), image_name)
            img_info = {
                "name": image_name,
                "meta": {
                    "deviceId": self.name,
                    "timestamp": timestamp,
                    "sensorsData": {
                        "extrinsicMatrix": list(self.extrinsic.flatten().astype(float)),
                        "intrinsicMatrix": list(self.intrinsic.flatten().astype(float)),
                    },
                },
            }
            return (sly_path_img, img_info)


@dataclass
class Sample:
    """
    A class to represent a sample from the NuScenes dataset.
    """
    timestamp_us: float
    lidar_path: str
    anns: List[AnnotationObject]
    cam_data: List[CamData]
    timestamp: str = field(init=False)

    def __post_init__(self):
        self.timestamp = datetime.utcfromtimestamp(self.timestamp_us / 1e6).isoformat()

    @staticmethod
    def generate_boxes(nuscenes, boxes: List) -> Generator:
        """
        Generate ground truth boxes for a given set of boxes.

        :param nuscenes: The nuScenes dataset instance
        :type nuscenes: NuScenes
        :param boxes: A list of boxes to generate ground truth for
        :type boxes: List
        :returns: A generator that yields tuples containing the ground truth box, name, and instance token.
        :rtype: generator
        """
        from nuscenes.utils.data_classes import Box  # pylint: disable=import-error

        boxes: List[Box] = boxes

        locs = np.array([b.center for b in boxes]).reshape(-1, 3)
        dims = np.array([b.wlh for b in boxes]).reshape(-1, 3)
        rots = np.array([b.orientation.yaw_pitch_roll[0] for b in boxes]).reshape(-1, 1)

        gt_boxes = np.concatenate([locs, dims, -rots - np.pi / 2], axis=1)
        names = np.array([b.name for b in boxes])
        instance_tokens = [nuscenes.get("sample_annotation", box.token) for box in boxes]

        yield from zip(gt_boxes, names, instance_tokens)

    def convert_lidar_to_supervisely(self) -> str:
        """
        Converts a LiDAR point cloud file to the Supervisely format and saves it as a .pcd file.

        :returns: The file path of the saved .pcd file.
        :rtype: str
        """
        import open3d as o3d  # pylint: disable=import-error

        bin_file = Path(self.lidar_path)
        save_path = str(bin_file.with_suffix(".pcd"))

        b = np.fromfile(bin_file, dtype=np.float32).reshape(-1, 5)
        points = b[:, 0:3]
        intensity = b[:, 3]
        ring_index = b[:, 4]
        intensity_fake_rgb = np.zeros((intensity.shape[0], 3))
        intensity_fake_rgb[:, 0] = (
            intensity  # red The intensity measures the reflectivity of the objects
        )
        intensity_fake_rgb[:, 1] = (
            ring_index  # green ring index is the index of the laser ranging from 0 to 31
        )
        try:
            pc = o3d.geometry.PointCloud(o3d.utility.Vector3dVector(points))
            pc.colors = o3d.utility.Vector3dVector(intensity_fake_rgb)
            o3d.io.write_point_cloud(save_path, pc)
        except Exception as e:
            logger.warning(f"Error converting lidar to supervisely format: {e}")
        return save_path



import os
from typing import List

import supervisely.convert.volume.sly.sly_volume_helper as sly_volume_helper
from supervisely.volume_annotation.volume_annotation import VolumeAnnotation
from supervisely import ProjectMeta, logger
from supervisely.project.volume_project import VolumeProject, VolumeDataset
from supervisely.convert.base_converter import AvailableVolumeConverters
from supervisely.convert.volume.volume_converter import VolumeConverter
from supervisely.io.fs import JUNK_FILES, get_file_ext, get_file_name
from supervisely.io.json import load_json_file
from supervisely.volume.volume import is_valid_ext as validate_volume_ext


class SLYVolumeConverter(VolumeConverter):
    """Imports Supervisely volume project format (NRRD/NIfTI + JSON annotations) from local folder."""

    def __str__(self) -> str:
        return AvailableVolumeConverters.SLY

    @property
    def ann_ext(self) -> str:
        return ".json"

    @property
    def key_file_ext(self) -> str:
        return ".json"

    def generate_meta_from_annotation(
        self, ann_path: str, meta: ProjectMeta
    ) -> ProjectMeta:
        meta = sly_volume_helper.get_meta_from_annotation(ann_path, meta)
        return meta

    def validate_ann_file(self, ann_path: str, meta: ProjectMeta) -> bool:
        try:
            ann_json = load_json_file(ann_path)
            for key in sly_volume_helper.SLY_VOLUME_ANN_KEYS:
                value = ann_json.get(key)
                if value is None:
                    return False
            ann = VolumeAnnotation.from_json(ann_json, meta)  # , KeyIdMap())
            return True
        except Exception as e:
            logger.warning(f"Failed to validate annotation: {repr(e)}")
            return False

    def validate_key_file(self, key_file_path: str) -> bool:
        try:
            self._meta = ProjectMeta.from_json(load_json_file(key_file_path))
            return True
        except Exception:
            return False

    def validate_format(self) -> bool:
        if self.read_sly_project(self._input_data):
            return True

        detected_ann_cnt = 0
        vol_list, stl_dict, ann_dict, mask_dict = [], {}, {}, {}
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                ext = get_file_ext(full_path)
                # if file == "key_id_map.json":
                #     key_id_map_json = load_json_file(full_path)
                #     self._key_id_map = KeyIdMap.from_dict(key_id_map_json)
                if file == "meta.json":
                    is_valid = self.validate_key_file(full_path)
                    if is_valid:
                        if self._meta is not None:
                            meta_json = load_json_file(full_path)
                            try:
                                self._meta = self._meta.merge(
                                    ProjectMeta.from_json(meta_json)
                                )
                            except Exception as e:
                                logger.warning(f"Failed to merge meta: {repr(e)}")
                        continue

                elif file in JUNK_FILES:  # add better check
                    continue
                elif ext == ".stl":
                    stl_dir = os.path.dirname(full_path)
                    stl_dirname = os.path.basename(stl_dir)
                    stl_dict[stl_dirname] = full_path
                elif ext in self.ann_ext:
                    ann_dict[file] = full_path
                elif ext == ".dcm":
                    return False
                elif root.endswith(".nrrd") and os.path.isdir(root):
                    mask_dir = os.path.dirname(full_path)
                    mask_dirname = os.path.basename(mask_dir)
                    mask_dict[mask_dirname] = mask_dir
                else:
                    try:
                        is_valid = validate_volume_ext(ext)
                        if is_valid:
                            vol_list.append(full_path)
                    except:
                        continue

        if self._meta is not None:
            meta = self._meta
        else:
            meta = ProjectMeta()

        # create Items
        self._items = []
        for vol_path in vol_list:
            item = self.Item(vol_path)
            ann_name = f"{item.name}.json"
            if ann_name in ann_dict:
                ann_path = ann_dict[ann_name]
                if self._meta is None:
                    meta = self.generate_meta_from_annotation(ann_path, meta)
                is_valid = self.validate_ann_file(ann_path, meta)
                if is_valid:
                    item.ann_data = ann_path
                    detected_ann_cnt += 1
            item.mask_dir = mask_dict.get(item.name)
            item.interpolation_dir = stl_dict.get(item.name)
            self._items.append(item)
        self._meta = meta
        return detected_ann_cnt > 0

    def to_supervisely(
        self,
        item: VolumeConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> VolumeAnnotation:
        """Convert to Supervisely format."""
        if meta is None:
            meta = self._meta

        if item.ann_data is None:
            return item.create_empty_annotation()

        try:
            ann_json = load_json_file(item.ann_data)
            if renamed_classes or renamed_tags:
                ann_json = sly_volume_helper.rename_in_json(ann_json, renamed_classes, renamed_tags)
            return VolumeAnnotation.from_json(ann_json, meta)  # , KeyIdMap())
        except Exception as e:
            logger.warning(f"Failed to read annotation: {repr(e)}")
            return item.create_empty_annotation()

    def read_sly_project(self, input_data: str) -> bool:
        try:
            project_fs = VolumeProject.read_single(input_data)
            self._meta = project_fs.meta
            self._items = []
            
            for dataset_fs in project_fs.datasets:
                dataset_fs: VolumeDataset

                for item_name in dataset_fs:
                    img_path, ann_path = dataset_fs.get_item_paths(item_name)
                    item = self.Item(
                        item_path=img_path,
                        ann_data=ann_path,
                        shape=None,
                        interpolation_dir=dataset_fs.get_interpolation_dir(item_name),
                        mask_dir=dataset_fs.get_mask_dir(item_name),
                    )
                    self._items.append(item)
            return True

        except Exception as e:
            logger.info(f"Failed to read Supervisely project: {repr(e)}")
            return False



from supervisely.app.widgets import Widget, DynamicWidget
from supervisely.app.widgets_context import JinjaWidgets
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.app.fastapi.utils import run_sync
from supervisely.app.content import DataJson, StateJson


class ReloadableArea(DynamicWidget):
    """Widget for dynamic content reloading without reloading the whole app UI."""

    def __init__(self, content: Widget = None, widget_id: str = None):
        """Initialize the ReloadableArea widget.

        :param content: Widget to set as content. Defaults to None.
        :type content: :class:`~supervisely.app.widgets.widget.Widget`, optional
        :param widget_id: The id of the widget. Defaults to None.
        :type widget_id: str, optional

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import ReloadableArea, Container, Button

                button_plus = Button("plus")
                buttons_container = Container(widgets=[button_plus])
                reloadable_area = ReloadableArea(content=buttons_container)
                button_minus = Button("minus")
                buttons_container._widgets.append(button_minus)
                reloadable_area.reload()
        """
        self._content = content
        super().__init__(widget_id=widget_id, file_path=__file__)

        script_path = "./sly/css/app/widgets/reloadable_area/script.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__] = script_path

    def get_json_data(self):
        return None

    def get_json_state(self):
        return None

    def set_content(self, content: Widget):
        """Replaces content of the ReloadableArea with new widget. Note: this method doesn't reload the widget.
        To reload the widget in UI use reload() function.

        :param content: new widget to be set as content
        :type content: :class:`~supervisely.app.widgets.widget.Widget`
        """
        self._content = content
        DataJson().send_changes()
        StateJson().send_changes()

    def reload(self):
        """Reloads the widget in UI."""
        DataJson().send_changes()
        StateJson().send_changes()

        html_content = self._content.to_html()
        run_sync(
            WebsocketManager().broadcast(
                {
                    "runAction": {
                        "action": f"rerender-template-{self.widget_id}",
                        "payload": {"template": html_content},
                    }
                }
            )
        )

    def hide(self):
        """Hides the content of the ReloadableArea."""
        self._content.hide()

    def show(self):
        """Shows the content of the ReloadableArea."""
        self._content.show()



from typing import List, Tuple

from supervisely._utils import abs_url, is_debug_with_sly_net, is_development
from supervisely.app.widgets import (
    Button,
    Card,
    CheckboxField,
    ClassesTable,
    Container,
    Text,
)
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle
from supervisely.nn.task_type import TaskType
from supervisely.nn.training.gui.model_selector import ModelSelector


class ClassesSelector:
    """TrainApp GUI component for selecting model classes."""
    title = "Classes Selector"
    description = "Select classes that will be used for training"
    lock_message = "Select previous step to unlock"

    def __init__(
        self,
        project_id: int,
        classes: list,
        model_selector: ModelSelector = None,
        app_options: dict = {},
    ):
        """
        :param project_id: Project ID.
        :type project_id: int
        :param classes: Pre-selected classes.
        :type classes: list
        :param model_selector: Optional ModelSelector.
        :type model_selector: ModelSelector
        :param app_options: App options dict.
        :type app_options: dict
        """
        # Init widgets
        self.model_selector = model_selector
        self.qa_stats_text = None
        self.classes_table = None
        self.validator_text = None
        self.button = None
        self.container = None
        self.card = None
        # -------------------------------- #

        self.display_widgets = []
        self.app_options = app_options

        # GUI Components
        if is_development() or is_debug_with_sly_net():
            qa_stats_link = abs_url(f"projects/{project_id}/stats/datasets")
        else:
            qa_stats_link = f"/projects/{project_id}/stats/datasets"
        self.qa_stats_text = Text(
            text=f"<i class='zmdi zmdi-chart-donut' style='color: #7f858e'></i> <a href='{qa_stats_link}' target='_blank'> <b> QA & Stats </b></a>"
        )

        models = model_selector.models
        task_types = [model["meta"]["task_type"] for model in models]
        task_types = list(set(task_types))
        allowed_types = []
        for task_type in task_types:
            if task_type.endswith("detection"):
                allowed_types.append(Rectangle)
            elif task_type.endswith("segmentation"):
                allowed_types.extend([Bitmap, Polygon])
            elif task_type == TaskType.POSE_ESTIMATION:
                allowed_types.append(GraphNodes)

        self.classes_table = ClassesTable(project_id=project_id, allowed_types=allowed_types)
        if len(classes) > 0:
            self.classes_table.select_classes(classes)
        else:
            self.classes_table.select_all()

        self.validator_text = Text("")
        self.validator_text.hide()

        self.convert_class_shapes_checkbox = CheckboxField(
            title="Convert classes to model task type",
            description="If enabled, classes with compatible shapes will be converted according to the model CV task type",
            checked=False,
        )

        self.button = Button("Select")
        self.display_widgets.extend(
            [
                self.qa_stats_text,
                self.classes_table,
                self.convert_class_shapes_checkbox,
                self.validator_text,
                self.button,
            ]
        )
        # -------------------------------- #

        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
            collapsable=self.app_options.get("collapsable", False),
        )
        self.card.lock()

    @property
    def widgets_to_disable(self) -> list:
        return [self.classes_table, self.convert_class_shapes_checkbox]

    def get_selected_classes(self) -> list:
        return self.classes_table.get_selected_classes()

    def set_classes(self, classes) -> None:
        self.classes_table.select_classes(classes)

    def select_all_classes(self) -> None:
        self.classes_table.select_all()

    def get_wrong_shape_classes(self, task_type: str) -> List[str]:
        allowed_shapes = {
            TaskType.OBJECT_DETECTION: {Rectangle},
            TaskType.INSTANCE_SEGMENTATION: {Bitmap},
            TaskType.SEMANTIC_SEGMENTATION: {Bitmap},
            TaskType.POSE_ESTIMATION: {GraphNodes},
            TaskType.PROMPTABLE_SEGMENTATION: {Bitmap},
            TaskType.INTERACTIVE_SEGMENTATION: {Bitmap},
            TaskType.PROMPT_BASED_OBJECT_DETECTION: {Rectangle},
        }

        if task_type not in allowed_shapes:
            return []

        selected_classes = self.get_selected_classes()
        wrong_shape_classes = []
        for class_name in selected_classes:
            obj_class = self.classes_table.project_meta.get_obj_class(class_name)

            from supervisely.annotation.obj_class import ObjClass

            obj_class: ObjClass
            if obj_class is None:
                continue
            if obj_class.geometry_type not in allowed_shapes[task_type]:
                wrong_shape_classes.append(class_name)

        return wrong_shape_classes

    def classify_incompatible_classes(
        self, task_type: str
    ) -> Tuple[List[str], List[str]]:
        """
        Rules:
        1) Detection – any shape can be converted, Rectangle is compatible.
        2) Instance/Semantic segmentation – only Bitmap and Polygon (need conversion to Bitmap) are allowed; other shapes are not convertible.
        3) Pose estimation – only GraphNodes are allowed; other shapes are not convertible.

        Returns:
        - convertible: List[str] – list of class names that can be converted to the task type
        - non_convertible: List[str] – list of class names that cannot be converted to the task type
        """
        selected_classes = self.get_selected_classes()
        convertible: List[str] = []
        non_convertible: List[str] = []
        for class_name in selected_classes:
            obj_class = self.classes_table.project_meta.get_obj_class(class_name)
            if obj_class is None:
                continue

            geo_cls = obj_class.geometry_type

            if task_type == TaskType.OBJECT_DETECTION:
                if geo_cls is Rectangle:
                    # already compatible
                    continue
                # Any other shape is converted to Rectangle (BBox)
                convertible.append(class_name)

            elif task_type in (
                TaskType.INSTANCE_SEGMENTATION,
                TaskType.SEMANTIC_SEGMENTATION,
            ):
                if geo_cls is Bitmap:
                    # already compatible (bitmap mask)
                    continue
                if geo_cls is Polygon:
                    # convertible to bitmap mask
                    convertible.append(class_name)
                    continue
                # Other shapes cannot be converted
                non_convertible.append(class_name)

            elif task_type == TaskType.POSE_ESTIMATION:
                if geo_cls is GraphNodes:
                    # already compatible
                    continue
                # Other shapes cannot be converted
                non_convertible.append(class_name)

            else:
                # Unknown task type – treat all shapes as compatible
                continue

        return convertible, non_convertible

    def validate_step(self) -> bool:
        # @TODO: Handle AnyShape classes
        self.validator_text.hide()
        task_type = (
            self.model_selector.get_selected_task_type()
            if self.model_selector
            else None
        )

        if len(self.classes_table.project_meta.obj_classes) == 0:
            self.validator_text.set(text="Project has no classes", status="error")
            self.validator_text.show()
            return False

        selected_classes = self.classes_table.get_selected_classes()
        n_classes = len(selected_classes)

        if n_classes == 0:
            self.validator_text.set(
                text="Please select at least one class", status="error"
            )
            self.validator_text.show()
            return False

        class_word = "class" if n_classes == 1 else "classes"
        message_parts = [f"Selected {n_classes} {class_word}"]
        status = "success"
        is_valid = True

        empty_classes = [
            row[0]["data"]
            for row in self.classes_table._table_data
            if row[0]["data"] in selected_classes
            and row[2]["data"] == 0
            and row[3]["data"] == 0
        ]
        if empty_classes:
            empty_word = "class" if len(empty_classes) == 1 else "classes"
            message_parts.append(
                f"{empty_word.capitalize()} with no annotations: {', '.join(empty_classes)}"
            )
            status = "warning"

        convertible_classes, non_convertible_classes = (
            self.classify_incompatible_classes(task_type)
        )

        if (
            n_classes > 0
            and not convertible_classes
            and len(non_convertible_classes) == n_classes
        ):
            available_task_types = []
            if self.model_selector is not None and hasattr(
                self.model_selector, "pretrained_models_table"
            ):
                available_task_types = (
                    self.model_selector.pretrained_models_table.get_available_task_types()
                )

            other_suitable_exists = False
            for other_task_type in available_task_types:
                if other_task_type == task_type:
                    continue
                other_conv, other_non_conv = self.classify_incompatible_classes(
                    other_task_type
                )
                if other_conv or len(other_non_conv) < n_classes:
                    other_suitable_exists = True
                    break

            if other_suitable_exists:
                self.validator_text.set(
                    text=(
                        f"No suitable classes for task type {task_type}. "
                        "Training is not possible. Please choose another task type."
                    ),
                    status="error",
                )
            else:
                self.validator_text.set(
                    text=(
                        "Project contains no suitable classes for training. "
                        "Training is not possible. Please select another project."
                    ),
                    status="error",
                )
            self.validator_text.show()
            return False

        incompatible_exist = bool(convertible_classes or non_convertible_classes)
        if incompatible_exist:
            task_specific_texts = {
                TaskType.OBJECT_DETECTION: "Only rectangle shapes are supported for object detection task, all other shape types can be converted to rectangle",
                TaskType.INSTANCE_SEGMENTATION: "Only bitmap shapes are supported for instance segmentation task, polygon shapes can be converted to bitmap",
                TaskType.SEMANTIC_SEGMENTATION: "Only bitmap shapes are supported for semantic segmentation task, polygon shapes can be converted to bitmap",
                TaskType.POSE_ESTIMATION: "Only keypoint (graph) shape is supported for pose estimation task",
            }

            if non_convertible_classes:
                specific_text = task_specific_texts.get(
                    task_type,
                    "Some selected classes have shapes that are incompatible with the chosen model task type.",
                )
                message_parts = [f"Model task type is {task_type}. {specific_text}"]
                message_parts.append("Remove incompatible classes to continue.")
                status = "error"
                is_valid = False
            else:
                if self.is_convert_class_shapes_enabled():
                    message_parts.append(
                        f"Conversion enabled. Classes will be converted for task {task_type}."
                    )
                    status = "info" if status == "success" else status
                    is_valid = True
                else:
                    specific_text = task_specific_texts.get(
                        task_type,
                        "Some selected classes have shapes that are incompatible with the chosen model task type.",
                    )
                    message_parts = [
                        f"Model task type is {task_type}. {specific_text}",
                        "Enable conversion checkbox or select compatible classes to continue.",
                    ]
                    status = "error"
                    is_valid = False
        else:
            if self.is_convert_class_shapes_enabled():
                message_parts.append(
                    "Conversion enabled, but no shape conversion required."
                )

        self.validator_text.set(text=". ".join(message_parts), status=status)
        self.validator_text.show()
        return is_valid

    def is_convert_class_shapes_enabled(self) -> bool:
        return self.convert_class_shapes_checkbox.is_checked()



from __future__ import annotations

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    MarkdownWidget,
    NotificationWidget,
)


class Recall(DetectionVisMetric):
    """Recall metrics and per-class recall chart for detection reports."""

    MARKDOWN = "recall"
    MARKDOWN_PER_CLASS = "recall_per_class"
    NOTIFICATION = "recall"
    CHART = "recall"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True

    @property
    def md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_R
        return MarkdownWidget(self.MARKDOWN, "Recall", text)

    @property
    def notification(self) -> NotificationWidget:
        title, desc = self.vis_texts.notification_recall.values()
        mp = self.eval_result.mp
        tp_plus_fn = mp.TP_count + mp.FN_count
        recall = mp.base_metrics()["recall"].round(2)
        if mp.average_across_iou_thresholds:
            iou_text = "[0.5,0.55,...,0.95]"
        else:
            if mp.iou_threshold_per_class is not None:
                iou_text = "custom"
            else:
                iou_text = mp.iou_threshold
        title = f"Recall (IoU={iou_text}) = {recall}"
        return NotificationWidget(
            self.NOTIFICATION,
            title,
            desc.format(mp.TP_count, tp_plus_fn),
        )

    @property
    def per_class_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_R_perclass.format(self.vis_texts.definitions.f1_score)
        return MarkdownWidget(self.MARKDOWN_PER_CLASS, "Recall per class", text)

    @property
    def chart(self) -> ChartWidget:
        chart = ChartWidget(self.CHART, self._get_figure())
        chart.set_click_data(
            self.explore_modal_table.id,
            self.get_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].label}`,",
        )
        return chart

    def _get_figure(self):  #  -> go.Figure
        import plotly.express as px  # pylint: disable=import-error

        sorted_by_f1 = self.eval_result.mp.per_class_metrics().sort_values(by="f1")
        fig = px.bar(
            sorted_by_f1,
            x="category",
            y="recall",
            color="recall",
            range_color=[0, 1],
            color_continuous_scale="Plasma",
        )
        fig.update_traces(hovertemplate="Class: %{x}<br>Recall: %{y:.2f}<extra></extra>")
        if len(sorted_by_f1) <= 20:
            fig.update_traces(
                text=sorted_by_f1["recall"].round(2),
                textposition="outside",
            )
        fig.update_xaxes(title_text="Class")
        fig.update_yaxes(title_text="Recall", range=[0, 1])
        fig.update_layout(
            width=700 if len(sorted_by_f1) < 10 else None,
        )
        return fig



from __future__ import annotations

import asyncio
import re
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Callable, List, Literal, Optional
from uuid import uuid4

from fastapi.responses import HTMLResponse
from pydantic import BaseModel

from supervisely._utils import is_production
from supervisely.api.api import Api
from supervisely.app.widgets import Widget
from supervisely.io.env import task_id as env_task_id


class DebouncedEventHandler:
    """Aggregate frequent events and call a handler once per debounce interval."""

    def __init__(self, debounce_time: float = 0.1):
        """
        :param debounce_time: Time in seconds to wait before calling handler.
        :type debounce_time: float
        """
        self._event_queue = []
        self._debounce_time = debounce_time
        self._task = None

    async def _process_events(self, func: Callable):
        await asyncio.sleep(self._debounce_time)
        aggregated_events = self._event_queue.copy()
        self._event_queue.clear()
        func(aggregated_events)

    def handle_event(self, event_data, func: Callable):
        self._event_queue.append(event_data)
        if self._task is None or self._task.done():
            self._task = asyncio.create_task(self._process_events(func))


class SelectedIds(BaseModel):
    """Pydantic model for selected point IDs returned from the Bokeh widget."""

    selected_ids: List[int]


class Bokeh(Widget):
    """
    Bokeh widget for creating interactive plots. Only Bokeh version 3.1.1 is supported.

    This widget uses Bokeh library to create interactive plots.
    """

    class Routes:
        """HTTP routes used by the widget to serve HTML and receive selection events."""

        VALUE_CHANGED = "value_changed"
        HTML_ROUTE = "bokeh.html"

    class Plot(ABC):
        """Base class for Bokeh plot layers that can be added to a figure with a shared data source."""

        def __init__(self, name: Optional[str] = None, **kwargs):
            """
            :param name: Name of the plot layer. Auto-generated UUID if None.
            :type name: str, optional
            :param kwargs: Additional kwargs passed to Bokeh plot methods.
            """
            self._name = name or str(uuid4())
            self.kwargs = kwargs

        @abstractmethod
        def add(self, plot, source) -> None:
            pass

        @property
        def name(self) -> str:
            return self._name

    class Circle(Plot):
        """Circle plot layer (grouped by `names` field in the data source)."""

        def add(self, plot, source) -> None:
            from bokeh.models import (  # pylint: disable=import-error
                CDSView,
                GroupFilter,
            )

            filters = [GroupFilter(column_name="names", group=self.name, name=self.name)]
            view = CDSView(source=source, filters=filters)
            return plot.circle(
                "x",
                "y",
                radius="radius",
                fill_color="colors",
                fill_alpha=0.5,
                source=source,
                line_color=None,
                view=view,
            )

    class Scatter(Plot):
        """Scatter plot layer (grouped by `names` field in the data source)."""

        def add(self, plot, source) -> None:
            from bokeh.models import (  # pylint: disable=import-error
                CDSView,
                GroupFilter,
            )

            filters = [GroupFilter(column_name="names", group=self.name, name=self.name)]
            view = CDSView(source=source, filters=filters)
            return plot.scatter(
                "x",
                "y",
                size="radius",
                color="colors",
                fill_alpha=0.5,
                source=source,
                view=view,
            )

    class Line(Plot):
        """Line plot layer (grouped by `names` field in the data source)."""

        def add(self, plot, source) -> None:
            from bokeh.models import (  # pylint: disable=import-error
                CDSView,
                GroupFilter,
            )

            filters = [GroupFilter(column_name="names", group=self.name, name=self.name)]
            view = CDSView(source=source, filters=filters)
            return plot.line("x", "y", source=source, view=view, line_width=2)

    def __init__(
        self,
        plots: List[Plot] = None,
        width: int = 1000,
        height: int = 600,
        tools: List[str] = [
            "pan",
            "wheel_zoom",
            "box_zoom",
            "reset",
            # "save",
            "poly_select",
            "tap",
            "lasso_select",
        ],
        toolbar_location: Literal["above", "below", "left", "right"] = "above",
        x_axis_visible: bool = False,
        y_axis_visible: bool = False,
        grid_visible: bool = False,
        widget_id: Optional[str] = None,
        show_legend: bool = False,
        legend_location: Literal["left", "top", "right", "bottom"] = "right",
        legend_click_policy: Literal["hide", "mute"] = "hide",
        **kwargs,
    ):
        """
        :param plots: List of plots to be displayed.
        :type plots: List[Plot]
        :param width: Width of the chart in pixels.
        :type width: int
        :param height: Height of the chart in pixels.
        :type height: int
        :param tools: List of tools to be displayed on the chart.
        :type tools: List[str]
        :param toolbar_location: Location of the toolbar.
        :type toolbar_location: Literal["above", "below", "left", "right"]
        :param x_axis_visible: If True, x-axis will be visible.
        :type x_axis_visible: bool
        :param y_axis_visible: If True, y-axis will be visible.
        :type y_axis_visible: bool
        :param grid_visible: If True, grid will be visible.
        :type grid_visible: bool
        :param widget_id: Unique widget identifier.
        :type widget_id: str
        :param show_legend: If True, clickable legend widget will be displayed.
        :type show_legend: bool
        :param legend_location: Location of the clickable legend widget.
        :type legend_location: Literal["left", "top", "right", "bottom"]
        :param legend_click_policy: Click policy of the clickable legend widget.
        :type legend_click_policy: Literal["hide", "mute"]

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import Bokeh, IFrame

                data = {
                    "x": [1, 2, 3, 4, 5], "y": [1, 2, 3, 4, 5],
                    "radius": [10, 20, 30, 40, 50],
                    "colors": ["red", "green", "blue", "yellow", "purple"],
                    "ids": [1, 2, 3, 4, 5], "names": ["kiwi", "kiwi", "lemon", "lemon", "lemon"],
                }
                plot_lemon = Bokeh.Circle(name="lemon")
                plot_kiwi = Bokeh.Circle(name="kiwi")
                bokeh = Bokeh(x_axis_visible=True, y_axis_visible=True, grid_visible=True, show_legend=True)
                bokeh.add_data(**data)
                bokeh.add_plots([plot_lemon, plot_kiwi])
                iframe = IFrame()
                iframe.set(bokeh.html_route_with_timestamp, height="650px", width="100%")
        """
        import bokeh  # pylint: disable=import-error

        # check Bokeh version compatibility (only 3.1.1 is supported)
        if bokeh.__version__ != "3.1.1":
            raise RuntimeError(f"Bokeh version {bokeh.__version__} is not supported. Use 3.1.1")

        self._source_data = {"x": [], "y": [], "radius": [], "colors": [], "ids": [], "names": []}
        self._source = None
        self.widget_id = widget_id
        self._plots = plots or []
        self._view = None

        self._width = width
        self._height = height
        self._tools = tools
        self._toolbar_location = toolbar_location
        self._x_axis_visible = x_axis_visible
        self._y_axis_visible = y_axis_visible
        self._grid_visible = grid_visible
        self._show_legend = show_legend
        self._legend_location = legend_location
        self._legend_click_policy = legend_click_policy

        super().__init__(widget_id=widget_id, file_path=__file__)

        self._update()

        server = self._sly_app.get_server()

        @server.get(self.html_route)
        def _html_response() -> None:
            return HTMLResponse(content=self._get_html())

    @property
    def html_route(self) -> str:
        return self.get_route_path(Bokeh.Routes.HTML_ROUTE)

    @property
    def html_route_with_timestamp(self) -> str:
        return f".{self.html_route}?t={datetime.now().timestamp()}"

    def get_json_data(self):
        return {}

    def get_json_state(self):
        return {}

    def _add_callbacks(self):
        from bokeh.models import CustomJS  # pylint: disable=import-error

        route_path = self.get_route_path(Bokeh.Routes.VALUE_CHANGED)

        if is_production():
            api = Api()
            task_info = api.task.get_info_by_id(env_task_id())
            if task_info is not None:
                route_path = f"/net/{task_info['meta']['sessionToken']}{route_path}"
        callback = CustomJS(
            args=dict(source=self._source),
            code=f"""
                var indices = source.selected.indices;
                var selected_ids = [];
                for (var i = 0; i < indices.length; i++) {{
                    selected_ids.push(source.data['ids'][indices[i]]);
                }}
                var xhr = new XMLHttpRequest();
                xhr.open("POST", "{route_path}", true);
                xhr.setRequestHeader("Content-Type", "application/json");
                xhr.send(JSON.stringify({{selected_ids: selected_ids}}));
            """,
        )
        self._source.selected.js_on_change("indices", callback)

    def _get_html(self) -> str:
        return f"""<div>
            <script type="text/javascript" src="https://cdn.bokeh.org/bokeh/release/bokeh-3.1.1.min.js"></script>
            {self._script}
            {self._div}
        </div>"""

    def _create_figure(self):
        from bokeh.models import (  # pylint: disable=import-error
            ColumnDataSource,
            Legend,
        )
        from bokeh.plotting import figure  # pylint: disable=import-error

        self._plot = figure(width=self._width, height=self._height, tools=self._tools)

        self._plot.xaxis.visible = self._x_axis_visible
        self._plot.yaxis.visible = self._y_axis_visible
        self._plot.grid.visible = self._grid_visible

        self._source = ColumnDataSource(data=self._source_data)
        self._add_callbacks()
        legend_items = []
        for plot in self._plots:
            r = plot.add(self._plot, self._source)
            legend_items.append((plot.name, [r]))

        if self._show_legend:
            self._plot.add_layout(
                Legend(items=legend_items, click_policy=self._legend_click_policy),
                self._legend_location,
            )

    def _update_html(self):
        from bokeh.embed import components  # pylint: disable=import-error

        script, self._div = components(self._plot)
        self._div_id = self._get_div_id(self._div)
        self._script = self._update_script(script)

    def _update(self):
        self._create_figure()
        self._update_html()

    def _get_div_id(self, div: str) -> str:
        match = re.search(r'id="([^"]+)"', div)
        if match:
            return match.group(1)
        raise ValueError(f"No div id found in {div}")

    def _update_script(self, script: str) -> str:
        insert_after = "const fn = function() {"
        updated_script = ""
        for line in script.split("\n"):
            if line.strip().startswith(insert_after):
                line += f"""
                    const el = document.querySelector('#{self._div_id}');
                    if (el === null) {{
                        setTimeout(fn, 200);
                        return;
                    }}
                """
            updated_script += line + "\n"
        return updated_script

    def value_changed(self, func: Callable) -> Callable:
        """Registers a callback function that will be called when the chart is clicked."""

        server = self._sly_app.get_server()
        route_path = self.get_route_path(Bokeh.Routes.VALUE_CHANGED)
        self._changes_handled = True
        debounced_handler = DebouncedEventHandler(debounce_time=0.2)

        @server.post(route_path)
        async def _click(data: SelectedIds) -> None:
            debounced_handler.handle_event(data.selected_ids, func)

        return _click

    def clear(self) -> None:
        """Clears all data in the ColumnDataSource and removes plots."""

        self._source_data = {key: [] for key in self._source_data.keys()}
        self._plots = []
        self._update()

    def refresh(self) -> None:
        """Refreshes the chart by reloading the existing data and updating the HTML."""

        self._update()

    def update_chart_size(self, width: Optional[int] = None, height: Optional[int] = None):
        """Updates the size of the chart."""

        if width:
            self._width = width
        if height:
            self._height = height
        self._update()

    def add_data(
        self,
        x: List[float],
        y: List[float],
        radius: List[float],
        colors: List[str],
        ids: List[Any],
        names: List[str],
        append: bool = True,
    ):
        """Adds data to the chart."""

        if append:
            self._source.data["x"] += x
            self._source.data["y"] += y
            self._source.data["radius"] += radius
            self._source.data["colors"] += colors
            self._source.data["ids"] += ids
            self._source.data["names"] += names
        else:
            self._source.data = {
                "x": x,
                "y": y,
                "radius": radius,
                "colors": colors,
                "ids": ids,
                "names": names,
            }

    def add_plot(self, plot: Plot):
        """Adds a plot to the chart."""

        self._plots.append(plot)
        self._update()

    def add_plots(self, plots: List[Plot]):
        """Adds multiple plots to the chart."""

        self._plots.extend(plots)
        self._update()

    def update_point_size(self, size: float):
        """Updates the size of the points in the chart."""

        self._source_data["radius"] = [size] * len(self._source_data["x"])
        self._update()



import traceback
from typing import Dict, List, Optional

from supervisely import logger
from supervisely.app import StateJson
from supervisely.app.content import DataJson
from supervisely.app.widgets import Widget

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Tabs(Widget):
    """Tabs widget that shows one of several content panes and supports click callbacks."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        CLICK = "tab_clicked_cb"

    class TabPane:
        """One tab pane (label + content widget)."""

        def __init__(self, label: str, content: Widget):
            """:param label: Tab label.
            :type label: str
            :param content: Widget to display when tab is active.
            :type content: Widget
            """
            self.label = label
            self.name = label  # identifier corresponding to the active tab
            self.content = content

    def __init__(
        self,
        labels: List[str],
        contents: List[Widget],
        type: Optional[Literal["card", "border-card"]] = "border-card",
        widget_id=None,
    ):
        """:param labels: List of tab labels (unique, max 10).
        :type labels: List[str]
        :param contents: List of widgets, one per tab.
        :type contents: List[Widget]
        :param type: Style: "card" or "border-card".
        :type type: Literal["card", "border-card"], optional
        :param widget_id: Unique widget identifier.

        :raises ValueError: If labels/contents lengths differ, or labels not unique.
        """
        if len(labels) != len(contents):
            raise ValueError("labels length must be equal to contents length in Tabs widget.")
        if len(labels) > 10:
            raise ValueError("You can specify up to 10 tabs.")
        if len(set(labels)) != len(labels):
            raise ValueError("All of tab labels should be unique.")
        self._items: List[Tabs.TabPane] = []
        for label, widget in zip(labels, contents):
            self._items.append(Tabs.TabPane(label=label, content=widget))
        self._value = labels[0]
        self._type = type
        self._click_handled = False
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {
            "type": self._type,
            "tabsOptions": {item.name: {"disabled": False} for item in self._items},
        }

    def get_json_state(self) -> Dict:
        return {"value": self._value}

    def set_active_tab(self, value: str):
        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_active_tab(self) -> str:
        return StateJson()[self.widget_id]["value"]

    def disable_tab(self, tab_name: str):
        if tab_name not in [item.name for item in self._items]:
            raise ValueError(f"Tab with name '{tab_name}' does not exist.")
        DataJson()[self.widget_id]["tabsOptions"][tab_name]["disabled"] = True
        DataJson().send_changes()

    def enable_tab(self, tab_name: str):
        if tab_name not in [item.name for item in self._items]:
            raise ValueError(f"Tab with name '{tab_name}' does not exist.")
        DataJson()[self.widget_id]["tabsOptions"][tab_name]["disabled"] = False
        DataJson().send_changes()

    def click(self, func):
        route_path = self.get_route_path(Tabs.Routes.CLICK)
        server = self._sly_app.get_server()

        self._click_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_active_tab()
            if res is not None:
                try:
                    return func(res)
                except Exception as e:
                    logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
                    raise e

        return _click



from typing import Literal, Optional

from supervisely._utils import abs_url, is_debug_with_sly_net, is_development
from supervisely.api.file_api import FileInfo
from supervisely.app import DataJson
from supervisely.app.widgets import Widget
from supervisely.imaging.color import _validate_hex_color


class ReportThumbnail(Widget):
    """Compact card for an evaluation report (benchmark or experiment) with link and custom icon colors."""

    def __init__(
        self,
        info: Optional[FileInfo] = None,
        widget_id: Optional[str] = None,
        title: Optional[str] = None,
        color: Optional[str] = "#dcb0ff",
        bg_color: Optional[str] = "#faebff",
        report_type: Literal["model_benchmark", "experiment"] = "model_benchmark",
    ):
        """:param info: FileInfo for report file. Can be set later.
        :type info: FileInfo, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        :param title: Custom title. Defaults to "Evaluation Report".
        :type title: str, optional
        :param color: Icon color (hex).
        :type color: str, optional
        :param bg_color: Icon background color (hex).
        :type bg_color: str, optional
        :param report_type: "model_benchmark" or "experiment".
        :type report_type: Literal["model_benchmark", "experiment"]
        """
        self._id: int = None
        self._info: FileInfo = None
        self._description: str = None
        self._url: str = None
        self._report_type = report_type
        self._set_info(info)
        self._title = title
        self._color = color if _validate_hex_color(color) else "#dcb0ff"
        self._bg_color = bg_color if _validate_hex_color(bg_color) else "#faebff"

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "id": self._id,
            "name": self._title or "Evaluation Report",
            "description": self._description,
            "url": self._url,
            "description": self._description,
            "icon": {
                "className": "zmdi zmdi-assignment",
                "color": self._color,
                "bgColor": self._bg_color,
            },
        }

    def get_json_state(self):
        return None

    def _set_info(self, info: FileInfo = None):
        if info is None:
            self._id: int = None
            self._info: FileInfo = None
            self._description: str = None
            self._url: str = None
            return
        self._id = info.id
        self._info = info
        if self._report_type == "model_benchmark":
            self._description = "Open the Model Benchmark evaluation report."
            lnk = f"/model-benchmark?id={info.id}"
        elif self._report_type == "experiment":
            self._description = "Open the Experiment report."
            lnk = f"/nn/experiments/{info.id}"
        else:
            raise ValueError(f"Invalid report type: {self._report_type}")

        lnk = abs_url(lnk) if is_development() or is_debug_with_sly_net() else lnk
        # self._description = info.path
        self._url = lnk

    def set(
        self,
        info: FileInfo = None,
        report_type: Optional[Literal["model_benchmark", "experiment"]] = None,
    ):
        if report_type is not None:
            self._report_type = report_type
        self._set_info(info)
        self.update_data()
        DataJson().send_changes()



# coding: utf-8

"""Work with point cloud episode annotations via the Supervisely API."""

from supervisely.api.entity_annotation.entity_annotation_api import EntityAnnotationAPI
from supervisely.api.module_api import ApiField
from supervisely.pointcloud_annotation.pointcloud_episode_annotation import (
    PointcloudEpisodeAnnotation,
)
from supervisely.video_annotation.key_id_map import KeyIdMap


class PointcloudEpisodeAnnotationAPI(EntityAnnotationAPI):
    """API for working with point cloud episode annotations."""

    _method_download = "point-clouds.episodes.annotations.info"
    _entity_ids_str = ApiField.POINTCLOUD_IDS

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                dataset_id = 62664
                ann_info = api.pointcloud_episode.annotation.download(dataset_id)
                print(ann_info)
        """
        super().__init__(api)

    def download(self, dataset_id: int) -> dict:
        """
        Download information about PointcloudEpisodeAnnotation by dataset ID from API.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :returns: Dictionary with information about PointcloudEpisodeAnnotation in json format.
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                dataset_id = 62664
                ann_info = api.pointcloud_episode.annotation.download(dataset_id)
                print(ann_info)

                # Output: {
                #     'datasetId': 62664,
                #     'description': '',
                #     'frames': [
                #         {'figures': [{'classId': None,
                #                     'createdAt': '2023-04-05T08:55:52.445Z',
                #                     'description': '',
                #                     'geometry': {'dimensions': {'x': 2.3652234,
                #                                                 'y': 23.291742,
                #                                                 'z': 3.326648},
                #                                 'position': {'x': 86.29707472161449,
                #                                             'y': -14.472597682830635,
                #                                             'z': 0.8842007608554671},
                #                                 'rotation': {'x': 0,
                #                                             'y': 0,
                #                                             'z': -1.6962800995995606}},
                #                     'geometryType': 'cuboid_3d',
                #                     'id': 87830452,
                #                     'labelerLogin': 'almaz',
                #                     'objectId': 5565741,
                #                     'updatedAt': '2023-04-05T08:55:52.445Z'},
                #                     {'classId': None,
                #                     'createdAt': '2023-04-05T08:55:52.445Z',
                #                     'description': '',
                #                     'geometry': {'indices': [783,
                #                                             784,
                #                                             ...
                #                                             28326,
                #                                             30294]},
                #                     'geometryType': 'point_cloud',
                #                     'id': 87830456,
                #                     'labelerLogin': 'almaz',
                #                     'objectId': 5565740,
                #                     'updatedAt': '2023-04-05T08:55:52.445Z'}],
                #         'index': 0,
                #         'pointCloudId': 19618654},...
                #     ],
                #     'tags': []
                # }
        """

        response = self._api.post(self._method_download, {ApiField.DATASET_ID: dataset_id})
        result = response.json()
        if len(result) == 0:
            return PointcloudEpisodeAnnotation().to_json()
        return response.json()[0]

    def download_bulk(self, dataset_id, entity_ids):
        """Not supported for point cloud episodes."""
        raise RuntimeError("Not supported for episodes")

    def append(
        self,
        dataset_id: int,
        ann: PointcloudEpisodeAnnotation,
        frame_to_pointcloud_ids,
        key_id_map: KeyIdMap = None,
    ) -> None:
        """
        Loads an PointcloudEpisodeAnnotation to a given point cloud ID in the API.

        :param pointcloud_id: Point cloud ID in Supervisely.
        :type pointcloud_id: int
        :param ann: PointcloudEpisodeAnnotation object.
        :type ann: :class:`~supervisely.pointcloud_annotation.pointcloud_episode_annotation.PointcloudEpisodeAnnotation`
        :param frame_to_pointcloud_ids:  List of dictionaries with frame_id and name of pointcloud episodes
        :type frame_to_pointcloud_ids: List[dict]
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                pcd_episodes_id = 198704259
                api.pointcloud_episode.annotation.append(pcd_episodes_id, pointcloud_ann)
        """

        if key_id_map is None:
            # create for internal purposes (to link figures and tags to objects)
            key_id_map = KeyIdMap()

        figures = []
        pointcloud_ids = []
        for frame in ann.frames:
            for fig in frame.figures:
                if frame_to_pointcloud_ids.get(frame.index) is None:  # skip unmapped frames
                    continue

                figures.append(fig)
                pointcloud_ids.append(frame_to_pointcloud_ids[frame.index])

        if len(pointcloud_ids) == 0:
            return
        self._api.pointcloud_episode.object.append_bulk(pointcloud_ids[0], ann.objects, key_id_map)
        self._api.pointcloud_episode.figure.append_to_dataset(
            dataset_id, figures, pointcloud_ids, key_id_map
        )



from supervisely.app.widgets import Widget
from supervisely.app import DataJson, StateJson
from supervisely import (
    TagMetaCollection,
    ObjClassCollection,
    TagMeta,
    ObjClass,
    TagValueType,
    color,
)
from typing import Union, List


class MatchTagMetasOrClasses(Widget):
    """Widget to compare and match TagMetas or ObjClasses between two collections (with optional suffix matching)."""

    def __init__(
        self,
        left_collection: Union[TagMetaCollection, ObjClassCollection, None] = None,
        right_collection: Union[TagMetaCollection, ObjClassCollection, None] = None,
        left_name: Union[str, None] = None,
        right_name: Union[str, None] = None,
        selectable: bool = False,
        suffix: Union[str, None] = None,
        widget_id: str = None,
    ):
        """
        :param left_collection: Left TagMetaCollection or ObjClassCollection.
        :type left_collection: Union[TagMetaCollection, ObjClassCollection, None]
        :param right_collection: Right collection (same type as left).
        :type right_collection: Union[TagMetaCollection, ObjClassCollection, None]
        :param left_name: Label for left column.
        :type left_name: str, optional
        :param right_name: Label for right column.
        :type right_name: str, optional
        :param selectable: If True, rows are selectable.
        :type selectable: bool
        :param suffix: Optional suffix for matching.
        :type suffix: str, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional

        :raises TypeError: If left_collection and right_collection types differ.
        """
        if not type(left_collection) is type(right_collection):
            raise TypeError("Collections should be of same type")
        self._collections_type = type(left_collection)
        self._left_collection = left_collection
        self._right_collection = right_collection
        if left_name is None:
            self._left_name = (
                "Left Tag Metas" if self._collections_type is TagMetaCollection else "Left Classes"
            )
        else:
            self._left_name = left_name
        if right_name is None:
            self._right_name = (
                "Right Tag Metas"
                if self._collections_type is TagMetaCollection
                else "Right Classes"
            )
        else:
            self._right_name = right_name
        self._selectable = selectable
        self._suffix = suffix

        self._table = self._get_table()

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "table": self._table,
            "left_name": self._left_name,
            "right_name": self._right_name,
            "selectable": self._selectable,
        }

    def get_json_state(self):
        return {"selected": []}

    def set(
        self,
        left_collection: Union[TagMetaCollection, ObjClassCollection, None] = None,
        right_collection: Union[TagMetaCollection, ObjClassCollection, None] = None,
        left_name: Union[str, None] = None,
        right_name: Union[str, None] = None,
        suffix: Union[str, None] = None,
    ):
        if not type(left_collection) is type(right_collection):
            raise TypeError("Collections should be of same type")
        self._collections_type = type(left_collection)
        self._left_collection = left_collection
        self._right_collection = right_collection
        self._left_name = left_name if left_name is not None else self._left_name
        self._right_name = right_name if right_name is not None else self._right_name
        self._suffix = suffix

        self._table = self._get_table()
        DataJson()[self.widget_id] = self.get_json_data()
        DataJson().send_changes()
        StateJson()[self.widget_id] = self.get_json_state()
        StateJson().send_changes()

    def get_stat(self):
        stat = {
            "match": [],
            "only_left": [],
            "only_right": [],
            "different_shape": [],
            "different_value_type": [],
            "different_one_of_options": [],
            "match_suffix": [],
            "different_shape_suffix": [],
            "different_value_type_suffix": [],
            "different_one_of_options_suffix": [],
        }
        for row in self._table:
            message_to_key = {
                "Match": "match",
                "Not found in right Project": "only_left",
                "Not found in left Project": "only_right",
                "Different shape": "different_shape",
                "Different value type": "different_value_type",
                "Type OneOf: conflict of possible values": "different_one_of_options",
                "Match with suffix": "match_suffix",
                "[Match with suffix] Different shape": "different_shape_suffix",
                "[Match with suffix] Different value type": "different_value_type_suffix",
                "[Match with suffix] Type OneOf: conflict of possible values": "different_one_of_options_suffix",
            }
            message = row.get("infoMessage")
            name1 = row.get("name1", None)
            name2 = row.get("name2", None)
            stat[message_to_key[message]].append((name1, name2))
        return stat

    def get_selected(self):
        return StateJson()[self.widget_id]["selected"]

    def _get_table(self):
        if self._left_collection is None:
            return []
        items1 = {item.name: 1 for item in self._left_collection}
        items2 = {item.name: 1 for item in self._right_collection}
        names = items1.keys() | items2.keys()
        mutual = items1.keys() & items2.keys()

        def get_mutual_with_suffix(names1, names2, suffix):
            left = {}
            right = {}
            l = len(suffix)
            for name1 in names1:
                name1: str
                if name1.endswith(suffix) and name1[:-l] in names2:
                    right[name1[:-l]] = name1
            for name2 in names2:
                name2: str
                if name2.endswith(suffix) and name2[:-l] in names1:
                    left[name2[:-l]] = name2
            return left, right

        mutual_with_suffix_left = {}
        mutual_with_suffix_right = {}
        if self._suffix is not None:
            mutual_with_suffix_left, mutual_with_suffix_right = get_mutual_with_suffix(
                items1.keys() - mutual, items2.keys() - mutual, self._suffix
            )

        diff1 = (
            items1.keys()
            - mutual
            - mutual_with_suffix_left.keys()
            - set(mutual_with_suffix_right.values())
        )
        diff2 = (
            items2.keys()
            - mutual
            - mutual_with_suffix_right.keys()
            - set(mutual_with_suffix_left.values())
        )

        match = []
        differ = []
        match_suffix = []
        differ_suffix = []
        missed = []

        def set_info(d, index, meta):
            d[f"name{index}"] = meta.name
            d[f"color{index}"] = color.rgb2hex(meta.color)
            if type(meta) is ObjClass:
                d[f"shape{index}"] = meta.geometry_type.geometry_name()
                d[f"shapeIcon{index}"] = "zmdi zmdi-shape"
            else:
                meta: TagMeta
                d[f"shape{index}"] = meta.value_type
                d[f"shapeIcon{index}"] = "zmdi zmdi-label"

        for name in names:
            compare = {}
            meta1 = self._left_collection.get(name)
            if meta1 is not None:
                set_info(compare, 1, meta1)
            meta2 = self._right_collection.get(name)
            if meta2 is not None:
                set_info(compare, 2, meta2)

            if name in mutual:
                flag = True
                diff_msg = ""
                if type(meta1) is ObjClass and meta1.geometry_type != meta2.geometry_type:
                    flag = False
                    diff_msg = "Different shape"
                if type(meta1) is TagMeta:
                    meta1: TagMeta
                    meta2: TagMeta
                    if meta1.value_type != meta2.value_type:
                        flag = False
                        diff_msg = "Different value type"
                    if meta1.value_type == TagValueType.ONEOF_STRING and set(
                        meta1.possible_values
                    ) != set(meta2.possible_values):
                        flag = False
                        diff_msg = "Type OneOf: conflict of possible values"

                if flag is False:
                    compare["infoMessage"] = diff_msg
                    compare["infoColor"] = "#FFBF00"
                    compare["infoIcon"] = (["zmdi zmdi-flag"],)
                    differ.append(compare)
                else:
                    compare["infoMessage"] = "Match"
                    compare["infoColor"] = "green"
                    compare["infoIcon"] = (["zmdi zmdi-check"],)
                    match.append(compare)
            elif name in mutual_with_suffix_left.keys() | mutual_with_suffix_right.keys():
                diff_msg = ""
                if name in mutual_with_suffix_left:
                    meta2 = self._right_collection.get(name + self._suffix)
                    set_info(compare, 2, meta2)
                else:
                    meta1 = self._left_collection.get(name + self._suffix)
                    set_info(compare, 1, meta1)
                flag = True
                if type(meta1) is ObjClass and meta1.geometry_type != meta2.geometry_type:
                    flag = False
                    diff_msg = "[Match with suffix] Different shape"
                if type(meta1) is TagMeta:
                    meta1: TagMeta
                    meta2: TagMeta
                    if meta1.value_type != meta2.value_type:
                        flag = False
                        diff_msg = "[Match with suffix] Different value type"
                    if meta1.value_type == TagValueType.ONEOF_STRING and set(
                        meta1.possible_values
                    ) != set(meta2.possible_values):
                        flag = False
                        diff_msg = "[Match with suffix] Type OneOf: conflict of possible values"

                if flag is False:
                    compare["infoMessage"] = diff_msg
                    compare["infoColor"] = "#FFBF00"
                    compare["infoIcon"] = (["zmdi zmdi-flag"],)
                    differ_suffix.append(compare)
                else:
                    compare["infoMessage"] = "Match with suffix"
                    compare["infoColor"] = "green"
                    compare["infoIcon"] = (["zmdi zmdi-check"],)
                    match_suffix.append(compare)
            elif name in diff1 | diff2:
                if name in diff1:
                    compare["infoMessage"] = "Not found in right Project"
                    compare["infoIcon"] = [
                        "zmdi zmdi-alert-circle-o",
                        "zmdi zmdi-long-arrow-right",
                    ]
                    compare["iconPosition"] = "right"
                else:
                    compare["infoMessage"] = "Not found in left Project"
                    compare["infoIcon"] = [
                        "zmdi zmdi-long-arrow-left",
                        "zmdi zmdi-alert-circle-o",
                    ]
                compare["infoColor"] = "red"
                missed.append(compare)

        table = []
        if match:
            match.sort(key=lambda x: x["name1"])
        table.extend(match)
        if differ:
            differ.sort(key=lambda x: x["name1"])
        table.extend(differ)
        if match_suffix:
            match_suffix.sort(key=lambda x: x["name1"])
        table.extend(match_suffix)
        if differ_suffix:
            differ_suffix.sort(key=lambda x: x["name1"])
        table.extend(differ_suffix)
        table.extend(missed)

        return table


class MatchTagMetas(MatchTagMetasOrClasses):
    """Specialization of :class:`MatchTagMetasOrClasses` for matching :class:`~supervisely.annotation.tag_meta.TagMeta`."""

    def __init__(
        self,
        left_collection: Union[TagMetaCollection, List[TagMeta], None] = None,
        right_collection: Union[TagMetaCollection, List[TagMeta], None] = None,
        left_name: Union[str, None] = None,
        right_name: Union[str, None] = None,
        selectable: bool = False,
        suffix: Union[str, None] = None,
        widget_id: str = None,
    ):
        """
        :param left_collection: Left TagMetaCollection or list of TagMeta.
        :type left_collection: Union[TagMetaCollection, List[TagMeta], None]
        :param right_collection: Right collection (same type as left).
        :type right_collection: Union[TagMetaCollection, List[TagMeta], None]
        :param left_name: Label for left column.
        :type left_name: str, optional
        :param right_name: Label for right column.
        :type right_name: str, optional
        :param selectable: If True, rows are selectable.
        :type selectable: bool
        :param suffix: Optional suffix for matching.
        :type suffix: str, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        if type(left_collection) is list:
            left_collection = TagMetaCollection(left_collection)
        if type(right_collection) is list:
            right_collection = TagMetaCollection(right_collection)

        super().__init__(
            left_collection=left_collection,
            right_collection=right_collection,
            left_name=left_name,
            right_name=right_name,
            selectable=selectable,
            suffix=suffix,
            widget_id=widget_id,
        )

    def set(
        self,
        left_collection: Union[TagMetaCollection, List[TagMeta], None] = None,
        right_collection: Union[TagMetaCollection, List[TagMeta], None] = None,
        left_name: Union[str, None] = None,
        right_name: Union[str, None] = None,
        suffix: Union[str, None] = None,
    ):
        if type(left_collection) is list:
            left_collection = TagMetaCollection(left_collection)
        if type(right_collection) is list:
            right_collection = TagMetaCollection(right_collection)

        super().set(
            left_collection=left_collection,
            right_collection=right_collection,
            left_name=left_name,
            right_name=right_name,
            suffix=suffix,
        )

    def get_stat(self):
        stat = super().get_stat()
        fields = {
            "match",
            "only_left",
            "only_right",
            "different_value_type",
            "different_one_of_options",
            "match_suffix",
            "different_value_type_suffix",
            "different_one_of_options_suffix",
        }
        stat = {key: value for key, value in stat.items() if key in fields}
        return stat


class MatchObjClasses(MatchTagMetasOrClasses):
    """Specialization of :class:`MatchTagMetasOrClasses` for matching :class:`~supervisely.annotation.obj_class.ObjClass`."""

    def __init__(
        self,
        left_collection: Union[ObjClassCollection, List[ObjClass], None] = None,
        right_collection: Union[ObjClassCollection, List[ObjClass], None] = None,
        left_name: Union[str, None] = None,
        right_name: Union[str, None] = None,
        selectable: bool = False,
        suffix: Union[str, None] = None,
        widget_id: str = None,
    ):
        """
        :param left_collection: Left ObjClassCollection or list of ObjClass.
        :type left_collection: Union[ObjClassCollection, List[ObjClass], None]
        :param right_collection: Right collection (same type as left).
        :type right_collection: Union[ObjClassCollection, List[ObjClass], None]
        :param left_name: Label for left column.
        :type left_name: str, optional
        :param right_name: Label for right column.
        :type right_name: str, optional
        :param selectable: If True, rows are selectable.
        :type selectable: bool
        :param suffix: Optional suffix for matching.
        :type suffix: str, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        if type(left_collection) is list:
            left_collection = ObjClassCollection(left_collection)
        if type(right_collection) is list:
            right_collection = ObjClassCollection(right_collection)

        super().__init__(
            left_collection=left_collection,
            right_collection=right_collection,
            left_name=left_name,
            right_name=right_name,
            selectable=selectable,
            suffix=suffix,
            widget_id=widget_id,
        )

    def set(
        self,
        left_collection: Union[ObjClassCollection, List[ObjClass], None] = None,
        right_collection: Union[ObjClassCollection, List[ObjClass], None] = None,
        left_name: Union[str, None] = None,
        right_name: Union[str, None] = None,
        suffix: Union[str, None] = None,
    ):
        if type(left_collection) is list:
            left_collection = ObjClassCollection(left_collection)
        if type(right_collection) is list:
            right_collection = ObjClassCollection(right_collection)

        super().set(
            left_collection=left_collection,
            right_collection=right_collection,
            left_name=left_name,
            right_name=right_name,
            suffix=suffix,
        )

    def get_stat(self):
        stat = super().get_stat()
        fields = {
            "match",
            "only_left",
            "only_right",
            "different_shape",
            "match_suffix",
            "different_shape_suffix",
        }
        stat = {key: value for key, value in stat.items() if key in fields}
        return stat



from typing import Callable, List, Optional, Union

from supervisely.api.user_api import UserInfo
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class SelectUser(Widget):
    """Dropdown for selecting users from a team. Extends the Select widget with user-specific functionality."""

    class Routes:
        """Route name constants for this widget."""
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        users: Optional[List[UserInfo]] = None,
        team_id: Optional[int] = None,
        roles: Optional[List[str]] = None,
        filterable: Optional[bool] = True,
        placeholder: Optional[str] = "Select user",
        size: Optional[Literal["large", "small", "mini"]] = None,
        multiple: bool = False,
        widget_id: Optional[str] = None,
    ):
        """Initialize the SelectUser widget.

        :param users: Initial list of :class:`~supervisely.api.user_api.UserInfo` instances
        :type users: Optional[List[:class:`~supervisely.api.user_api.UserInfo`]]
        :param team_id: Team ID to fetch users from
        :type team_id: Optional[int]
        :param roles: List of allowed user roles to filter by (e.g., ['admin', 'developer'])
        :type roles: Optional[List[str]]
        :param filterable: Enable search/filter functionality in dropdown
        :type filterable: Optional[bool]
        :param placeholder: Placeholder text when no user is selected
        :type placeholder: Optional[str]
        :param size: Size of the select dropdown
        :type size: Optional[Literal["large", "small", "mini"]]
        :param multiple: Enable multiple selection
        :type multiple: bool
        :param widget_id: Unique widget identifier
        :type widget_id: Optional[str]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.app.widgets import SelectUser

                select_user = SelectUser(team_id=123, roles=['admin', 'developer'], multiple=True)
                # Or: select_user = SelectUser(roles=['annotator']); select_user.set_users(user_list)
                @select_user.value_changed
                def on_user_selected(users):
                    if isinstance(users, list):
                        print(f"Selected users: {[u.login for u in users]}")
                    else:
                        print(f"Selected user: {users.login}")
        """
        self._users = []
        self._team_id = team_id
        self._allowed_roles = roles
        self._filterable = filterable
        self._placeholder = placeholder
        self._size = size
        self._multiple = multiple
        self._changes_handled = False

        # Load users from team_id if provided
        if team_id is not None:
            self._load_users_from_team(team_id)
        elif users is not None:
            self._users = self._filter_users_by_role(list(users))

        # Initialize parent Widget
        super().__init__(widget_id=widget_id, file_path=__file__)

    def _filter_users_by_role(self, users: List[UserInfo]) -> List[UserInfo]:
        """Filter users by allowed roles."""
        if self._allowed_roles is None:
            return users

        return [user for user in users if user.role in self._allowed_roles]

    def _load_users_from_team(self, team_id: int):
        """Load users from a team using the API."""
        from supervisely import Api

        api = Api.from_env()
        all_users = api.user.get_team_members(team_id)
        self._users = self._filter_users_by_role(all_users)

    def get_json_data(self):
        """Build JSON data for the widget."""
        items = []
        for user in self._users:
            user_name = None
            if user.name:
                name = user.name.strip()
                if name:
                    user_name = name[:15] + "…" if len(name) > 15 else name

            right_text = ""
            user_role = user.role.upper() if user.role else "NONE"
            if len(user_role) > 10:
                user_role = user_role[:10] + "…"
            if user_name and user_name != user.login:
                right_text = f"{user_name} • {user_role}"
            else:
                right_text = user_role

            items.append(
                {
                    "value": user.login,
                    "label": user.login,
                    "rightText": right_text,
                }
            )

        return {
            "items": items,
            "placeholder": self._placeholder,
            "filterable": self._filterable,
            "multiple": self._multiple,
            "size": self._size,
        }

    def get_json_state(self):
        """Build JSON state for the widget."""
        value = None
        if self._multiple:
            value = []
        return {"value": value}

    def get_value(self) -> Union[str, List[str], None]:
        """Get the currently selected user login(s)."""
        return StateJson()[self.widget_id]["value"]

    def get_selected_user(self) -> Union[UserInfo, List[UserInfo], None]:
        """Get the currently selected UserInfo object(s)."""
        value = self.get_value()
        if value is None:
            return None

        if self._multiple:
            if not isinstance(value, list):
                return []
            result = []
            for login in value:
                for user in self._users:
                    if user.login == login:
                        result.append(user)
                        break
            return result
        else:
            for user in self._users:
                if user.login == value:
                    return user
            return None

    def set_value(self, login: Union[str, List[str]]):
        """Set the selected user by login."""
        StateJson()[self.widget_id]["value"] = login
        StateJson().send_changes()

    def get_all_users(self) -> List[UserInfo]:
        """Get all available users."""
        return self._users.copy()

    def set_users(self, users: List[UserInfo]):
        """Update the list of available users."""
        self._users = self._filter_users_by_role(list(users))

        # Update data
        DataJson()[self.widget_id] = self.get_json_data()
        DataJson().send_changes()

        # Reset value if current selection is not in new users
        current_value = StateJson()[self.widget_id]["value"]
        if current_value:
            if self._multiple:
                if isinstance(current_value, list):
                    # Keep only valid selections
                    valid = [v for v in current_value if any(u.login == v for u in self._users)]
                    if valid != current_value:
                        StateJson()[self.widget_id]["value"] = valid
                        StateJson().send_changes()
            else:
                if not any(u.login == current_value for u in self._users):
                    StateJson()[self.widget_id]["value"] = (
                        self._users[0].login if self._users else None
                    )
                    StateJson().send_changes()

    def set_team_id(self, team_id: int):
        """Load users from a team by team_id."""
        self._team_id = team_id
        self._load_users_from_team(team_id)

        # Update data
        DataJson()[self.widget_id] = self.get_json_data()
        DataJson().send_changes()

        # Reset selection
        if self._multiple:
            StateJson()[self.widget_id]["value"] = []
        else:
            StateJson()[self.widget_id]["value"] = self._users[0].login if self._users else None
        StateJson().send_changes()

    def set_selected_users_by_ids(self, user_ids: Union[int, List[int]]):
        """Set the selected user(s) by user ID(s).

        :param user_ids: Single user ID or list of user IDs to select
        :type user_ids: Union[int, List[int]]
        """
        if isinstance(user_ids, int):
            user_ids = [user_ids]

        # Find logins for the given user IDs
        selected_logins = []
        for user_id in user_ids:
            for user in self._users:
                if user.id == user_id:
                    selected_logins.append(user.login)
                    break

        # Set value based on multiple mode
        if self._multiple:
            StateJson()[self.widget_id]["value"] = selected_logins
        else:
            StateJson()[self.widget_id]["value"] = selected_logins[0] if selected_logins else None

        StateJson().send_changes()

    def value_changed(self, func: Callable[[Union[UserInfo, List[UserInfo]]], None]):
        """
        Decorator to handle value change event.
        The decorated function receives the selected UserInfo (or list of UserInfo if multiple=True).

        :param func: Function to be called when selection changes
        :type func: Callable[[Union[:class:`~supervisely.api.user_api.UserInfo`, List[:class:`~supervisely.api.user_api.UserInfo`]]], None]
        """
        route_path = self.get_route_path(SelectUser.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _value_changed():
            selected = self.get_selected_user()
            if selected is not None:
                func(selected)

        return _value_changed



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from typing import List, Union


class Rate(Widget):
    """Star rating widget (optionally half-precision) that emits value-changed callbacks."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        value: Union[int, float] = None,
        max: int = 5,
        disabled: bool = False,
        allow_half: bool = False,
        texts: List[str] = [],
        show_text: bool = False,
        text_color: str = "#1F2D3D",
        text_template: str = "",
        colors: List = ["#F7BA2A", "#F7BA2A", "#F7BA2A"],
        void_color: str = "#C6D1DE",
        disabled_void_color: str = "#EFF2F7",
        widget_id: str = None,
    ):
        """:param value: Initial rating value.
        :type value: Union[int, float], optional
        :param max: Maximum stars (default 5).
        :type max: int
        :param disabled: If True, widget is disabled.
        :type disabled: bool
        :param allow_half: If True, allow half-star ratings.
        :type allow_half: bool
        :param texts: Optional text labels for each level.
        :type texts: List[str]
        :param show_text: If True, show text.
        :type show_text: bool
        :param text_color: Text color.
        :type text_color: str
        :param text_template: Template for disabled state.
        :type text_template: str
        :param colors: List of star colors.
        :param void_color: Color for unselected stars.
        :type void_color: str
        :param disabled_void_color: Color when disabled.
        :type disabled_void_color: str
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._value = value
        self._max = max
        self._disabled = disabled
        self._allow_half = allow_half

        # text properties
        self._texts = texts
        self._show_text = show_text
        self._text_color = text_color
        self._text_template = text_template  # available if self._disabled is True

        # color properties
        self._colors = colors
        self._void_color = void_color
        self._disabled_void_color = disabled_void_color


        self._changes_handled = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "max": self._max,
            "disabled": self._disabled,
            "allowHalf": self._allow_half,
            "texts": self._texts,
            "showText": self._show_text,
            "textColor": self._text_color,
            "textTemplate": self._text_template,
            "colors": self._colors,
            "voidColor": self._void_color,
            "disabledVoidColor": self._disabled_void_color,
        }

    def get_json_state(self):
        return {"value": self._value}

    @property
    def is_disabled(self):
        self._disabled = DataJson()[self.widget_id]["disabled"]
        return self._disabled

    def get_value(self):
        self._value = StateJson()[self.widget_id]["value"]
        return self._value

    def set_value(self, value: Union[int, float]):
        if not isinstance(value, (int, float)):
            raise TypeError("Value type has to be one of int or float.")
        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_max_value(self):
        self._max = DataJson()[self.widget_id]["max"]
        return self._max

    def set_max_value(self, value: int):
        if not isinstance(value, int):
            raise TypeError("Max value type has to be a integer.")
        self._max = value
        DataJson()[self.widget_id]["max"] = self._max
        DataJson().send_changes()

    def get_colors(self):
        self._colors = DataJson()[self.widget_id]["colors"]
        return self._colors

    def set_colors(self, value: List[str]):
        if not isinstance(value, list):
            raise TypeError("Argument value type has to be a list.")
        self._colors = value
        DataJson()[self.widget_id]["colors"] = self._colors
        DataJson().send_changes()

    def allow_half_precision(self):
        self._allow_half = True
        DataJson()[self.widget_id]["allow_half"] = self._allow_half
        DataJson().send_changes()

    def disallow_half_precision(self):
        self._allow_half = False
        DataJson()[self.widget_id]["allow_half"] = self._allow_half
        DataJson().send_changes()

    def get_texts(self):
        self._texts = DataJson()[self.widget_id]["texts"]
        return self._texts

    def set_texts(self, value: List[str]):
        if not isinstance(value, list):
            raise TypeError("Argument value type has to be a list.")
        self._texts = value
        DataJson()[self.widget_id]["texts"] = self._texts
        DataJson().send_changes()

    def show_text(self):
        self._show_text = True
        DataJson()[self.widget_id]["show_text"] = self._show_text
        DataJson().send_changes()

    def hide_text(self):
        self._show_text = False
        DataJson()[self.widget_id]["show_text"] = self._show_text
        DataJson().send_changes()

    def get_text_color(self):
        self._text_color = DataJson()[self.widget_id]["text_color"]
        return self._text_color

    def set_text_color(self, value: str):
        if not isinstance(value, str):
            raise TypeError("Argument value type has to be a str.")
        self._text_color = value
        DataJson()[self.widget_id]["text_color"] = self._text_color
        DataJson().send_changes()

    def get_void_color(self):
        self._void_color = DataJson()[self.widget_id]["void_color"]
        return self._void_color

    def set_void_color(self, value: str):
        if not isinstance(value, str):
            raise TypeError("Argument value type has to be a str.")
        self._void_color = value
        DataJson()[self.widget_id]["void_color"] = self._void_color
        DataJson().send_changes()

    def get_disabled_void_color(self):
        self._disabled_void_color = DataJson()[self.widget_id]["disabled_void_color"]
        return self._disabled_void_color

    def set_disabled_void_color(self, value: str):
        if not isinstance(value, str):
            raise TypeError("Argument value type has to be a str.")
        self._disabled_void_color = value
        DataJson()[self.widget_id]["disabled_void_color"] = self._disabled_void_color
        DataJson().send_changes()

    def value_changed(self, func):
        route_path = self.get_route_path(Rate.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            self.get_value()
            func(self._value)

        return _click



import asyncio
import concurrent.futures


def run_sync(coroutine):
    """
    Runs an asynchronous coroutine in a separate thread and waits for its result.
    It is useful for running async functions in a synchronous
    environment.

    This method creates a new thread using ThreadPoolExecutor and executes the coroutine
    inside a new event loop.

    ⚠️ Note: This function creates a new event loop every time it is called,
    which can cause issues when using objects tied to a specific loop (e.g., asyncio.Semaphore).

    :param coroutine: coroutine to run
    :returns: result of coroutine
    """
    try:
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            result = executor.submit(
                lambda coroutine_to_exec: asyncio.run(coroutine_to_exec), coroutine
            ).result()
        return result
    except RuntimeError as ex:
        print(repr(ex))
        return None



from supervisely.nn.benchmark.semantic_segmentation.base_vis_metric import (
    SemanticSegmVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class IntersectionErrorOverUnion(SemanticSegmVisMetric):
    """IoU and error-over-union bar charts for semantic segmentation reports."""


    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            "intersection_error_over_union",
            "Intersection & Error Over Union",
            text=self.vis_texts.markdown_iou,
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget("intersection_error_over_union", self.get_figure())

    def get_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        # Intersection & Error Over Union figure
        labels = ["mIoU", "mBoundaryEoU", "mExtentEoU", "mSegmentEoU"]
        values = [
            self.eval_result.mp.iou,
            self.eval_result.mp.boundary_eou,
            self.eval_result.mp.extent_eou,
            self.eval_result.mp.segment_eou,
        ]
        fig.add_trace(
            go.Pie(
                labels=labels,
                values=values,
                hole=0.5,
                textposition="outside",
                textinfo="percent+label",
                marker=dict(colors=["#8ACAA1", "#FFE4B5", "#F7ADAA", "#dd3f3f"]),
            )
        )

        return fig
